# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from botocore.awsrequest import AWSResponse
from botocore.compat import json
from botocore.exceptions import ClientError

import awscli.customizations.emr.createdefaultroles as createdefaultroles
import awscli.customizations.emr.emrutils as emrutils
from awscli.testutils import mock, unittest
from tests.unit.customizations.emr import (
    EMRBaseAWSCommandParamsTest as BaseAWSCommandParamsTest,
)

EC2_ROLE_NAME = "EMR_EC2_DefaultRole"
EMR_ROLE_NAME = "EMR_DefaultRole"
EMR_AUTOSCALING_ROLE_NAME = "EMR_AutoScaling_DefaultRole"

EC2_ROLE_POLICY = {
    "Statement": [
        {
            "Action": [
                "cloudwatch:*",
                "dynamodb:*",
                "ec2:Describe*",
                "elasticmapreduce:Describe*",
                "rds:Describe*",
                "s3:*",
                "sdb:*",
                "sns:*",
                "sqs:*",
            ],
            "Effect": "Allow",
            "Resource": ["*"],
        }
    ]
}

CREATE_EC2_ROLE_RESULT = {
    "Role": {
        "AssumeRolePolicyDocument": {
            "Version": "2008-10-17",
            "Statement": [
                {
                    "Action": "sts:AssumeRole",
                    "Sid": "",
                    "Effect": "Allow",
                    "Principal": {"Service": "ec2.amazonaws.com"},
                }
            ],
        },
        "RoleId": "AROAJG7O4RNNSRINMF6DI",
        "CreateDate": "2014-05-01T23:47:14.552Z",
        "RoleName": EC2_ROLE_NAME,
        "Path": "/",
        "Arn": "arn:aws:iam::176430881729:role/" + EC2_ROLE_NAME,
    }
}

CONSTRUCTED_RESULT_OUTPUT = [
    {"Role": CREATE_EC2_ROLE_RESULT['Role'], "RolePolicy": EC2_ROLE_POLICY}
]

http_response = AWSResponse(None, 200, {}, None)

CN_EC2_ROLE_ARN = (
    'arn:aws-cn:iam::aws:policy/service-role/'
    'AmazonElasticMapReduceforEC2Role'
)
US_GOV_EC2_ROLE_ARN = (
    'arn:aws-us-gov:iam::aws:policy/service-role/'
    'AmazonElasticMapReduceforEC2Role'
)

EC2_ROLE_ARN = (
    'arn:aws:iam::aws:policy/service-role/' 'AmazonElasticMapReduceforEC2Role'
)

CN_EMR_ROLE_ARN = (
    'arn:aws-cn:iam::aws:policy/service-role/' 'AmazonElasticMapReduceRole'
)

US_GOV_EMR_ROLE_ARN = (
    'arn:aws-us-gov:iam::aws:policy/' 'service-role/AmazonElasticMapReduceRole'
)

EMR_ROLE_ARN = (
    'arn:aws:iam::aws:policy/service-role/' 'AmazonElasticMapReduceRole'
)

CN_EMR_AUTOSCALING_ROLE_ARN = 'arn:aws-cn:iam::aws:policy/service-role/AmazonElasticMapReduceforAutoScalingRole'

US_GOV_EMR_AUTOSCALING_ROLE_ARN = (
    'arn:aws-us-gov:iam::aws:policy/service-role/'
    'AmazonElasticMapReduceforAutoScalingRole'
)

EMR_AUTOSCALING_ROLE_ARN = 'arn:aws:iam::aws:policy/service-role/AmazonElasticMapReduceforAutoScalingRole'


class TestCreateDefaultRole(BaseAWSCommandParamsTest):
    prefix = 'emr create-default-roles'

    ec2_role_policy_document = {
        "Version": "2008-10-17",
        "Statement": [
            {
                "Sid": "",
                "Effect": "Allow",
                "Principal": {"Service": "ec2.amazonaws.com"},
                "Action": "sts:AssumeRole",
            }
        ],
    }

    emr_role_policy_document = {
        "Version": "2008-10-17",
        "Statement": [
            {
                "Sid": "",
                "Effect": "Allow",
                "Principal": {"Service": "elasticmapreduce.amazonaws.com.cn"},
                "Action": "sts:AssumeRole",
            }
        ],
    }

    emr_autoscaling_role_policy_document_cn = {
        "Version": "2008-10-17",
        "Statement": [
            {
                "Sid": "",
                "Effect": "Allow",
                "Principal": {
                    "Service": [
                        "elasticmapreduce.amazonaws.com.cn",
                        "application-autoscaling.amazonaws.com.cn",
                    ]
                },
                "Action": "sts:AssumeRole",
            }
        ],
    }

    emr_autoscaling_role_policy_document = {
        "Version": "2008-10-17",
        "Statement": [
            {
                "Sid": "",
                "Effect": "Allow",
                "Principal": {
                    "Service": [
                        "elasticmapreduce.amazonaws.com",
                        "application-autoscaling.amazonaws.com",
                    ]
                },
                "Action": "sts:AssumeRole",
            }
        ],
    }

    def test_default_roles_exist(self):
        cmdline = self.prefix

        self.run_cmd(cmdline, expected_rc=0)
        self.assertEqual(len(self.operations_called), 4)

        self.assertEqual(self.operations_called[0][0].name, 'GetRole')
        self.assertEqual(
            self.operations_called[0][1]['RoleName'], EC2_ROLE_NAME
        )

        self.assertEqual(
            self.operations_called[1][0].name, 'GetInstanceProfile'
        )
        self.assertEqual(
            self.operations_called[1][1]['InstanceProfileName'], EC2_ROLE_NAME
        )
        self.assertEqual(self.operations_called[2][0].name, 'GetRole')
        self.assertEqual(
            self.operations_called[2][1]['RoleName'], EMR_ROLE_NAME
        )

        self.assertEqual(self.operations_called[3][0].name, 'GetRole')
        self.assertEqual(
            self.operations_called[3][1]['RoleName'], EMR_AUTOSCALING_ROLE_NAME
        )

    @mock.patch(
        'awscli.customizations.emr.emr.' 'CreateDefaultRoles._construct_result'
    )
    @mock.patch(
        'awscli.customizations.emr.emr.'
        'CreateDefaultRoles.check_if_instance_profile_exists'
    )
    @mock.patch(
        'awscli.customizations.emr.emr.'
        'CreateDefaultRoles.check_if_role_exists'
    )
    @mock.patch(
        'awscli.customizations.emr.emr.' 'CreateDefaultRoles._get_role_policy'
    )
    def test_default_autoscaling_role_commercial(
        self,
        get_rp_patch,
        role_exists_patch,
        instance_profile_exists_patch,
        construct_result_patch,
    ):
        get_rp_patch.return_value = False
        role_exists_patch.return_value = False
        instance_profile_exists_patch.return_value = False
        construct_result_patch.return_value = []

        cmdline = self.prefix + ' --region us-east-1'

        self.run_cmd(cmdline, expected_rc=0)

        # Only 8 operations will be called as we are mocking
        # check_if_role_exists and check_if_instance_profile_exists methods.
        self.assertEqual(len(self.operations_called), 8)
        self.assertEqual(
            self.operations_called[6][1]['AssumeRolePolicyDocument'],
            emrutils.dict_to_string(self.emr_autoscaling_role_policy_document),
        )

    @mock.patch(
        'awscli.customizations.emr.emr.' 'CreateDefaultRoles._construct_result'
    )
    @mock.patch(
        'awscli.customizations.emr.emr.'
        'CreateDefaultRoles.check_if_instance_profile_exists'
    )
    @mock.patch(
        'awscli.customizations.emr.emr.'
        'CreateDefaultRoles.check_if_role_exists'
    )
    @mock.patch(
        'awscli.customizations.emr.emr.' 'CreateDefaultRoles._get_role_policy'
    )
    def test_default_roles_not_exist(
        self,
        get_rp_patch,
        role_exists_patch,
        instance_profile_exists_patch,
        construct_result_patch,
    ):
        get_rp_patch.return_value = False
        role_exists_patch.return_value = False
        instance_profile_exists_patch.return_value = False
        construct_result_patch.return_value = []

        cmdline = self.prefix + ' --region cn-north-1'

        self.run_cmd(cmdline, expected_rc=0)

        # Only 8 operations will be called as we are mocking
        # check_if_role_exists and check_if_instance_profile_exists methods.
        self.assertEqual(len(self.operations_called), 8)

        self.assertEqual(self.operations_called[0][0].name, 'CreateRole')
        self.assertEqual(
            self.operations_called[0][1]['RoleName'], EC2_ROLE_NAME
        )
        self.assertEqual(
            self.operations_called[0][1]['AssumeRolePolicyDocument'],
            emrutils.dict_to_string(self.ec2_role_policy_document),
        )

        self.assertEqual(self.operations_called[1][0].name, 'AttachRolePolicy')
        self.assertEqual(
            self.operations_called[1][1]['PolicyArn'], CN_EC2_ROLE_ARN
        )
        self.assertEqual(
            self.operations_called[1][1]['RoleName'], EC2_ROLE_NAME
        )

        self.assertEqual(
            self.operations_called[2][0].name, 'CreateInstanceProfile'
        )
        self.assertEqual(
            self.operations_called[2][1]['InstanceProfileName'], EC2_ROLE_NAME
        )

        self.assertEqual(
            self.operations_called[3][0].name, 'AddRoleToInstanceProfile'
        )
        self.assertEqual(
            self.operations_called[3][1]['InstanceProfileName'], EC2_ROLE_NAME
        )
        self.assertEqual(
            self.operations_called[3][1]['RoleName'], EC2_ROLE_NAME
        )

        self.assertEqual(self.operations_called[4][0].name, 'CreateRole')
        self.assertEqual(
            self.operations_called[4][1]['RoleName'], EMR_ROLE_NAME
        )
        self.assertEqual(
            self.operations_called[4][1]['AssumeRolePolicyDocument'],
            emrutils.dict_to_string(self.emr_role_policy_document),
        )

        self.assertEqual(self.operations_called[5][0].name, 'AttachRolePolicy')
        self.assertEqual(
            self.operations_called[5][1]['PolicyArn'], CN_EMR_ROLE_ARN
        )
        self.assertEqual(
            self.operations_called[5][1]['RoleName'], EMR_ROLE_NAME
        )

        self.assertEqual(self.operations_called[6][0].name, 'CreateRole')
        self.assertEqual(
            self.operations_called[6][1]['RoleName'], EMR_AUTOSCALING_ROLE_NAME
        )
        self.assertEqual(
            self.operations_called[6][1]['AssumeRolePolicyDocument'],
            emrutils.dict_to_string(
                self.emr_autoscaling_role_policy_document_cn
            ),
        )

        self.assertEqual(self.operations_called[7][0].name, 'AttachRolePolicy')
        self.assertEqual(
            self.operations_called[7][1]['PolicyArn'],
            CN_EMR_AUTOSCALING_ROLE_ARN,
        )
        self.assertEqual(
            self.operations_called[7][1]['RoleName'], EMR_AUTOSCALING_ROLE_NAME
        )

    @mock.patch(
        'awscli.customizations.emr.emr.' 'CreateDefaultRoles._construct_result'
    )
    @mock.patch(
        'awscli.customizations.emr.createdefaultroles' '.get_service_principal'
    )
    @mock.patch(
        'awscli.customizations.emr.emr.'
        'CreateDefaultRoles.check_if_instance_profile_exists'
    )
    @mock.patch(
        'awscli.customizations.emr.emr.'
        'CreateDefaultRoles.check_if_role_exists'
    )
    @mock.patch(
        'awscli.customizations.emr.emr.' 'CreateDefaultRoles._get_role_policy'
    )
    def test_get_service_principal_parameters(
        self,
        get_rp_patch,
        role_exists_patch,
        instance_profile_exists_patch,
        get_sp_patch,
        construct_result_patch,
    ):
        get_rp_patch.return_value = "blah"
        get_sp_patch.return_value = 'elasticmapreduce.amazonaws.abc'
        role_exists_patch.return_value = False
        instance_profile_exists_patch.return_value = False
        construct_result_patch.return_value = []

        endpoint_url = 'https://elasticmapreduce.abc'
        cmdline = self.prefix + ' --endpoint ' + endpoint_url
        self.run_cmd(cmdline, expected_rc=0)
        self.assertEqual(get_sp_patch.call_args[0][1], endpoint_url)

    @mock.patch('botocore.session.Session.create_client')
    def test_call_parameters(self, call_patch):
        cmdline = self.prefix + ' --region eu-west-1' + ' --no-verify-ssl'
        self.run_cmd(cmdline, expected_rc=0)
        self.assertEqual(call_patch.call_args[1]['region_name'], 'eu-west-1')
        self.assertEqual(call_patch.call_args[1]['verify'], False)

    @mock.patch('botocore.session.Session.create_client')
    def test_call_parameters_only_endpoint(self, call_patch):
        endpoint_arg = 'https://elasticmapreduce.us-unknown-1.amazonaws.com'
        cmdline = self.prefix + ' --endpoint ' + endpoint_arg
        self.run_cmd(cmdline, expected_rc=0)
        self.assertEqual(call_patch.call_args[1]['endpoint_url'], None)

    @mock.patch('botocore.session.Session.create_client')
    def test_call_parameters_only_iam_endpoint(self, call_patch):
        endpoint_arg = 'https://elasticmapreduce.us-unknown-1.amazonaws.com'
        cmdline = self.prefix + ' --iam-endpoint ' + endpoint_arg
        self.run_cmd(cmdline, expected_rc=0)
        self.assertEqual(call_patch.call_args[1]['endpoint_url'], endpoint_arg)

    @mock.patch(
        'awscli.customizations.emr.emr.' 'CreateDefaultRoles._get_role_policy'
    )
    @mock.patch(
        'awscli.customizations.emr.emr.'
        'CreateDefaultRoles._create_role_with_role_policy'
    )
    @mock.patch(
        'awscli.customizations.emr.emr.'
        'CreateDefaultRoles.check_if_instance_profile_exists'
    )
    @mock.patch(
        'awscli.customizations.emr.emr.'
        'CreateDefaultRoles.check_if_role_exists'
    )
    def test_constructed_result(
        self,
        role_exists_patch,
        instance_profile_exists_patch,
        create_role_patch,
        get_role_policy_patch,
    ):
        role_exists_patch.side_effect = side_effect_ofcheck_if_role_exists
        instance_profile_exists_patch.return_value = False
        create_role_patch.return_value = CREATE_EC2_ROLE_RESULT
        get_role_policy_patch.return_value = EC2_ROLE_POLICY

        cmdline = self.prefix + ' --region cn-north-1'
        result = self.run_cmd(cmdline, 0)
        expected_output = (
            json.dumps(CONSTRUCTED_RESULT_OUTPUT, indent=4) + '\n'
        )
        self.assertEqual(result[0], expected_output)

    def test_policy_arn_construction(self):
        self.assertEqual(
            createdefaultroles.get_role_policy_arn(
                "cn-north-1", createdefaultroles.EC2_ROLE_POLICY_NAME
            ),
            CN_EC2_ROLE_ARN,
        )
        self.assertEqual(
            createdefaultroles.get_role_policy_arn(
                "us-gov-west-1", createdefaultroles.EC2_ROLE_POLICY_NAME
            ),
            US_GOV_EC2_ROLE_ARN,
        )
        self.assertEqual(
            createdefaultroles.get_role_policy_arn(
                "eu-west-1", createdefaultroles.EC2_ROLE_POLICY_NAME
            ),
            EC2_ROLE_ARN,
        )
        self.assertEqual(
            createdefaultroles.get_role_policy_arn(
                "cn-north-1", createdefaultroles.EMR_ROLE_POLICY_NAME
            ),
            CN_EMR_ROLE_ARN,
        )
        self.assertEqual(
            createdefaultroles.get_role_policy_arn(
                "us-gov-west-1", createdefaultroles.EMR_ROLE_POLICY_NAME
            ),
            US_GOV_EMR_ROLE_ARN,
        )
        self.assertEqual(
            createdefaultroles.get_role_policy_arn(
                "eu-west-1", createdefaultroles.EMR_ROLE_POLICY_NAME
            ),
            EMR_ROLE_ARN,
        )
        self.assertEqual(
            createdefaultroles.get_role_policy_arn(
                "cn-north-1",
                createdefaultroles.EMR_AUTOSCALING_ROLE_POLICY_NAME,
            ),
            CN_EMR_AUTOSCALING_ROLE_ARN,
        )
        self.assertEqual(
            createdefaultroles.get_role_policy_arn(
                "us-gov-west-1",
                createdefaultroles.EMR_AUTOSCALING_ROLE_POLICY_NAME,
            ),
            US_GOV_EMR_AUTOSCALING_ROLE_ARN,
        )
        self.assertEqual(
            createdefaultroles.get_role_policy_arn(
                "eu-west-1",
                createdefaultroles.EMR_AUTOSCALING_ROLE_POLICY_NAME,
            ),
            EMR_AUTOSCALING_ROLE_ARN,
        )


class TestCreateDefaultRoles(unittest.TestCase):
    def setUp(self):
        self.session = mock.Mock()
        self.client = mock.Mock()
        self.session.create_client.return_value = self.client
        self.command = createdefaultroles.CreateDefaultRoles(self.session)
        setattr(self.command, 'iam_endpoint_url', 'https://www.amazonaws.com')
        self.parsed_globals = mock.Mock()
        self.parsed_globals.verify_ssl = True

    def testcheck_if_role_exists_raises_client_error(self):
        error_response = {'Error': {'Code': 'foo'}}
        error = ClientError(error_response, 'GetRole')
        self.client.get_role.side_effect = error

        with self.assertRaises(ClientError):
            self.command.check_if_role_exists('role', self.parsed_globals)

    def test_check_role_not_found(self):
        error_response = {'Error': {'Code': 'NoSuchEntity'}}
        error = ClientError(error_response, 'GetRole')
        self.client.get_role.side_effect = error
        self.assertFalse(
            self.command.check_if_role_exists('role', self.parsed_globals)
        )

    def test_check_instance_profile_exists_raises_client_error(self):
        error_response = {'Error': {'Code': 'foo'}}
        error = ClientError(error_response, 'GetInstanceProfile')
        self.client.get_instance_profile.side_effect = error

        with self.assertRaises(ClientError):
            self.command.check_if_instance_profile_exists(
                'role', self.parsed_globals
            )

    def test_check_instance_profile_not_found(self):
        error_response = {'Error': {'Code': 'NoSuchEntity'}}
        error = ClientError(error_response, 'GetInstanceProfile')
        self.client.get_instance_profile.side_effect = error
        self.assertFalse(
            self.command.check_if_instance_profile_exists(
                'role', self.parsed_globals
            )
        )


def side_effect_ofcheck_if_role_exists(*args, **kwargs):
    if args[0] == EC2_ROLE_NAME:
        return False
    else:
        return True


if __name__ == "__main__":
    unittest.main()



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import os

import pytest
from botocore.paginate import Paginator
from botocore.session import Session
from ruamel.yaml import YAML

from awscli.customizations.configure.writer import ConfigFileWriter
from awscli.customizations.wizard import core, ui
from awscli.customizations.wizard.app import WizardValues
from awscli.testutils import mock, temporary_file, unittest


def load_wizard(yaml_str):
    yaml = YAML(typ="rt")
    data = yaml.load(yaml_str)
    return data


class FakeWizardValues(WizardValues):
    def __init__(self):
        self.values = {}
        self._values = self.values

    def __getitem__(self, item):
        if item not in self.values:
            self.values[item] = item
        return self.values[item]


class FakePrompter:
    def __init__(self, responses):
        self.responses = responses
        self.recorded_prompts = []

    def prompt(self, text, choices=None):
        response = self.responses.get(text)
        if choices is not None:
            entry = text, response, choices
        else:
            entry = text, response
        self.recorded_prompts.append(entry)
        return response


class TestPlanner(unittest.TestCase):
    def setUp(self):
        self.responses = {}
        self.prompter = FakePrompter(self.responses)
        self.planner = core.Planner(
            step_handlers={
                'static': core.StaticStep(),
                'prompt': core.PromptStep(self.prompter),
                'yesno-prompt': core.YesNoPrompt(self.prompter),
                'template': core.TemplateStep(),
            }
        )

    def test_can_prompt_for_single_value(self):
        loaded = load_wizard("""
        plan:
          start:
            values:
              name:
                type: prompt
                description: Enter user name
        """)

        self.responses['Enter user name'] = 'admin'
        parameters = self.planner.plan(loaded['plan'])
        self.assertEqual(parameters['name'], 'admin')

    def test_can_prompt_for_multiple_values_in_order(self):
        loaded = load_wizard("""
        plan:
          start:
            values:
              name:
                type: prompt
                description: Enter user name
              group:
                type: prompt
                description: Enter group name
        """)
        self.responses['Enter user name'] = 'myname'
        self.responses['Enter group name'] = 'wheel'

        parameters = self.planner.plan(loaded['plan'])
        self.assertEqual(parameters['name'], 'myname')
        self.assertEqual(parameters['group'], 'wheel')
        # We should also have prompted in the order that the keys
        # were defined.
        self.assertEqual(
            self.prompter.recorded_prompts,
            [('Enter user name', 'myname'), ('Enter group name', 'wheel')],
        )

    def test_can_prompt_for_conditional_values_true(self):
        loaded = load_wizard("""
        plan:
          start:
            values:
              should_stop:
                type: prompt
                description: Should we stop
            next_step:
              switch: should_stop
              yes: DONE
              no: ask_name
          ask_name:
            values:
              name:
                type: prompt
                description: Enter user name
        """)
        self.responses['Should we stop'] = 'no'
        self.responses['Enter user name'] = 'admin'

        parameters = self.planner.plan(loaded['plan'])
        self.assertEqual(parameters['should_stop'], 'no')
        self.assertEqual(parameters['name'], 'admin')
        self.assertEqual(
            self.prompter.recorded_prompts,
            [('Should we stop', 'no'), ('Enter user name', 'admin')],
        )

    def test_can_prompt_for_conditional_values_false(self):
        loaded = load_wizard("""
        plan:
          start:
            values:
              should_stop:
                type: prompt
                description: Should we stop
            next_step:
              switch: should_stop
              yes: DONE
              no: ask_name
          ask_name:
            values:
              name:
                type: prompt
                description: Enter user name
        """)
        self.responses['Should we stop'] = 'yes'

        parameters = self.planner.plan(loaded['plan'])
        self.assertEqual(parameters['should_stop'], 'yes')
        self.assertNotIn('name', parameters)
        self.assertEqual(
            self.prompter.recorded_prompts,
            [('Should we stop', 'yes')],
        )

    def test_can_prompt_with_choices_for_prompt(self):
        loaded = load_wizard("""
        plan:
          start:
            values:
              name:
                type: prompt
                description: Enter user name
                choices:
                    - display: Administrator
                      actual_value: admin
                    - display: Developer
                      actual_value: dev
        """)
        self.responses['Enter user name'] = 'admin'

        parameters = self.planner.plan(loaded['plan'])
        self.assertEqual(parameters['name'], 'admin')
        self.assertEqual(
            self.prompter.recorded_prompts,
            [
                (
                    'Enter user name',
                    'admin',
                    [
                        {'display': 'Administrator', 'actual_value': 'admin'},
                        {'display': 'Developer', 'actual_value': 'dev'},
                    ],
                )
            ],
        )

    def test_can_prompt_with_types(self):
        loaded = load_wizard("""
        plan:
          start:
            values:
              name:
                type: prompt
                description: Enter a string
                datatype: str
              age:
                type: prompt
                description: Enter an int
                datatype: int
              float:
                type: prompt
                description: Enter a float
                datatype: float
              trueval:
                type: prompt
                description: Enter a true value
                datatype: bool
              falseval:
                type: prompt
                description: Enter a false value
                datatype: bool
        """)
        # The responses will always return string values.
        self.responses['Enter a string'] = 'myname'
        self.responses['Enter an int'] = '100'
        self.responses['Enter a float'] = '2.0'
        self.responses['Enter a true value'] = 'true'
        self.responses['Enter a false value'] = 'false'

        parameters = self.planner.plan(loaded['plan'])
        self.assertEqual(parameters['name'], 'myname')
        self.assertEqual(parameters['age'], 100)
        self.assertEqual(parameters['float'], 2.0)
        self.assertEqual(parameters['trueval'], True)
        self.assertEqual(parameters['falseval'], False)

    def test_special_step_done_stops_run(self):
        loaded = load_wizard("""
        plan:
          start:
            values:
              foo:
                type: prompt
                description: Foo
            next_step: DONE
          # This step will never be executed.
          never_used:
            values:
              bar:
                type: prompt
                description: Bar
        """)
        self.responses['Foo'] = 'foo-value'

        parameters = self.planner.plan(loaded['plan'])
        self.assertEqual(parameters['foo'], 'foo-value')
        self.assertNotIn('bar', parameters)
        self.assertEqual(
            self.prompter.recorded_prompts,
            # We never prompt for the 'bar' value.
            [
                ('Foo', 'foo-value'),
            ],
        )

    def test_can_run_template_step(self):
        loaded = load_wizard("""
        plan:
          start:
            values:
              foo:
                type: prompt
                description: Foo
              bar:
                type: template
                value: "template-{foo}"
        """)
        self.responses['Foo'] = 'foo-value'

        parameters = self.planner.plan(loaded['plan'])
        self.assertEqual(parameters['foo'], 'foo-value')
        self.assertEqual(parameters['bar'], 'template-foo-value')
        self.assertEqual(
            self.prompter.recorded_prompts,
            [
                ('Foo', 'foo-value'),
            ],
        )

    def test_can_run_apicall_step(self):
        loaded = load_wizard("""
        plan:
          start:
            values:
              foo:
                type: apicall
                operation: iam.ListPolicies
                params:
                  Scope: AWS
        """)
        mock_session = mock.Mock(spec=Session)
        mock_client = mock.Mock()
        mock_session.create_client.return_value = mock_client
        mock_client.list_policies.return_value = {
            'Policies': ['foo'],
        }
        api_step = core.APICallStep(
            api_invoker=core.APIInvoker(session=mock_session)
        )
        planner = core.Planner(
            step_handlers={
                'apicall': api_step,
            },
        )
        parameters = planner.plan(loaded['plan'])
        self.assertEqual(parameters['foo'], {'Policies': ['foo']})

    def test_can_run_apicall_step_with_cache(self):
        loaded = load_wizard("""
        plan:
          start:
            values:
              foo:
                type: apicall
                operation: iam.ListPolicies
                params:
                  Scope: AWS
                cache: true
              use_cached_foo:
                type: apicall
                operation: iam.ListPolicies
                params:
                  Scope: AWS
                cache: true
        """)
        mock_session = mock.Mock(spec=Session)
        mock_client = mock.Mock()
        mock_session.create_client.return_value = mock_client
        mock_client.list_policies.return_value = {
            'Policies': ['foo'],
        }
        api_step = core.APICallStep(
            api_invoker=core.APIInvoker(session=mock_session)
        )
        planner = core.Planner(
            step_handlers={
                'apicall': api_step,
            },
        )
        parameters = planner.plan(loaded['plan'])
        self.assertEqual(
            parameters,
            {
                'foo': {'Policies': ['foo']},
                'use_cached_foo': {'Policies': ['foo']},
            },
        )
        self.assertEqual(mock_client.list_policies.call_count, 1)

    def test_can_run_apicall_step_with_paginate(self):
        loaded = load_wizard("""
        plan:
          start:
            values:
              all_policies:
                type: apicall
                operation: iam.ListPolicies
                params:
                  Scope: All
                paginate: true
        """)
        mock_session = mock.Mock(spec=Session)
        mock_client = mock.Mock()
        mock_paginator = mock.Mock(Paginator)
        mock_session.create_client.return_value = mock_client
        mock_client.can_paginate.return_value = True
        mock_client.get_paginator.return_value = mock_paginator
        mock_paginator.paginate.return_value.build_full_result.return_value = {
            'Policies': ['foo'],
        }
        api_step = core.APICallStep(
            api_invoker=core.APIInvoker(session=mock_session)
        )
        planner = core.Planner(
            step_handlers={
                'apicall': api_step,
            },
        )
        parameters = planner.plan(loaded['plan'])
        self.assertEqual(
            parameters,
            {
                'all_policies': {'Policies': ['foo']},
            },
        )
        mock_client.get_paginator.assert_called_with('list_policies')
        mock_paginator.paginate.assert_called_with(Scope='All')

    def test_can_run_apicall_step_with_query(self):
        loaded = load_wizard("""
        plan:
          start:
            values:
              foo:
                type: apicall
                operation: iam.ListPolicies
                params:
                  Scope: AWS
                query: Policies[].Name
        """)
        mock_session = mock.Mock(spec=Session)
        mock_client = mock.Mock()
        mock_session.create_client.return_value = mock_client
        mock_client.list_policies.return_value = {
            'Policies': [{'Name': 'one'}, {'Name': 'two'}],
        }
        api_step = core.APICallStep(
            api_invoker=core.APIInvoker(session=mock_session)
        )
        planner = core.Planner(
            step_handlers={
                'apicall': api_step,
            },
        )
        parameters = planner.plan(loaded['plan'])
        # Note this value is the result is applying the
        # Polices[].Name jmespath query to the response.
        self.assertEqual(parameters['foo'], ['one', 'two'])

    def test_can_use_static_value(self):
        loaded = load_wizard("""
        plan:
          start:
            values:
              foo:
                type: static
                value: myvalue
        """)
        parameters = self.planner.plan(loaded['plan'])
        self.assertEqual(parameters['foo'], 'myvalue')

    def test_can_use_static_value_as_non_string_type(self):
        loaded = load_wizard("""
        plan:
          start:
            values:
              foo:
                type: static
                value: [1, 2, 3]
        """)
        parameters = self.planner.plan(loaded['plan'])
        self.assertEqual(parameters['foo'], [1, 2, 3])

    def test_choices_can_be_variable_reference(self):
        loaded = load_wizard("""
        plan:
          start:
            values:
              choices_var:
                type: static
                value:
                  - display: Administrator
                    actual_value: admin
                  - display: Developer
                    actual_value: dev
              name:
                type: prompt
                description: Enter user name
                choices: choices_var
        """)
        self.responses['Enter user name'] = 'admin'

        parameters = self.planner.plan(loaded['plan'])
        self.assertEqual(parameters['name'], 'admin')
        self.assertEqual(
            self.prompter.recorded_prompts,
            [
                (
                    'Enter user name',
                    'admin',
                    [
                        {'display': 'Administrator', 'actual_value': 'admin'},
                        {'display': 'Developer', 'actual_value': 'dev'},
                    ],
                )
            ],
        )

    def test_can_run_fileprompt_step(self):
        loaded = load_wizard("""
        plan:
          start:
            values:
              foo:
                type: fileprompt
                description: Select file
        """)
        prompter = mock.Mock(spec=ui.UIFilePrompter)
        prompter.prompt.return_value = 'myfile.txt'
        planner = core.Planner(
            step_handlers={
                'fileprompt': core.FilePromptStep(prompter=prompter),
            },
        )
        parameters = planner.plan(loaded['plan'])
        self.assertEqual(parameters['foo'], os.path.abspath('myfile.txt'))

    def test_can_run_yes_no_prompt_step(self):
        loaded = load_wizard("""
        plan:
          test:
            values:
              wants_defaults:
                type: yesno-prompt
                question: "Do you want to use the defaults?"
        """)
        self.responses['Do you want to use the defaults?'] = 'yes'

        parameters = self.planner.plan(loaded['plan'])
        self.assertEqual(parameters['wants_defaults'], 'yes')
        self.assertEqual(
            self.prompter.recorded_prompts,
            [
                (
                    'Do you want to use the defaults?',
                    'yes',
                    [
                        {'display': 'Yes', 'actual_value': 'yes'},
                        {'display': 'No', 'actual_value': 'no'},
                    ],
                )
            ],
        )

    def test_can_set_default_yes_no_value(self):
        loaded = load_wizard("""
        plan:
          test:
            values:
              wants_defaults:
                type: yesno-prompt
                question: "Do you want to use the defaults?"
                start_value: no
        """)
        self.responses['Do you want to use the defaults?'] = 'no'

        parameters = self.planner.plan(loaded['plan'])
        self.assertEqual(
            self.prompter.recorded_prompts[0][2],
            # The default is No so it should be presented first.
            [
                {'display': 'No', 'actual_value': 'no'},
                {'display': 'Yes', 'actual_value': 'yes'},
            ],
        )

    def test_can_jump_around_to_next_steps(self):
        # This test shows that you can specify an explicit
        # next step name to jump to.
        loaded = load_wizard("""
        plan:
          step_a:
            values:
              first:
                type: prompt
                description: step_a
            next_step: step_d
          step_b:
            values:
              fourth:
                type: prompt
                description: step_b
            next_step: DONE
          step_c:
            values:
              third:
                type: prompt
                description: step_c
            next_step: step_b
          step_d:
            values:
              second:
                type: prompt
                description: step_d
            next_step: step_c
        """)
        # Note the order here, we should run the steps
        # in this order: step_a, step_d, step_c, step_b
        self.responses['step_a'] = 'one'
        self.responses['step_d'] = 'two'
        self.responses['step_c'] = 'three'
        self.responses['step_b'] = 'four'

        parameters = self.planner.plan(loaded['plan'])
        self.assertEqual(parameters['first'], 'one')
        self.assertEqual(parameters['second'], 'two')
        self.assertEqual(parameters['third'], 'three')
        self.assertEqual(parameters['fourth'], 'four')
        self.assertEqual(
            self.prompter.recorded_prompts,
            [
                ('step_a', 'one'),
                ('step_d', 'two'),
                ('step_c', 'three'),
                ('step_b', 'four'),
            ],
        )

    def test_can_delegate_to_arbitrary_type(self):
        class CustomStep(core.BaseStep):
            def run_step(self, step_definition, parameters):
                # Just return whatever the value of 'foo' is in the
                # step definition.
                return step_definition['foo']

        custom_step = CustomStep()
        custom_planner = core.Planner(
            step_handlers={
                'customstep': custom_step,
            },
        )
        loaded = load_wizard("""
        plan:
          start:
            values:
              name:
                type: customstep
                foo: myreturnvalue
        """)
        parameters = custom_planner.plan(loaded['plan'])
        self.assertEqual(parameters['name'], 'myreturnvalue')

    def test_can_load_profiles(self):
        loaded = load_wizard("""
        plan:
          start:
            values:
              foo:
                type: sharedconfig
                operation: ListProfiles
        """)
        config_api = mock.Mock(spec=core.SharedConfigAPI)
        config_api.list_profiles.return_value = ['profile1', 'profile2']
        sharedconfig = core.SharedConfigStep(
            config_api=config_api,
        )
        planner = core.Planner(
            step_handlers={
                'sharedconfig': sharedconfig,
            },
        )
        parameters = planner.plan(loaded['plan'])
        self.assertEqual(parameters['foo'], ['profile1', 'profile2'])

    def test_can_read_config_profile_data(self):
        loaded = load_wizard("""
        plan:
          start:
            values:
              foo:
                type: sharedconfig
                operation: GetValue
                params:
                  profile: devprofile
                  value: region
        """)
        config_api = mock.Mock(spec=core.SharedConfigAPI)
        config_api.get_value.return_value = 'us-west-2'
        sharedconfig = core.SharedConfigStep(
            config_api=config_api,
        )
        planner = core.Planner(
            step_handlers={
                'sharedconfig': sharedconfig,
            },
        )
        parameters = planner.plan(loaded['plan'])
        self.assertEqual(parameters['foo'], 'us-west-2')
        config_api.get_value.assert_called_with(
            profile='devprofile', value='region'
        )


class TestExecutor(unittest.TestCase):
    def setUp(self):
        self.session = mock.Mock(spec=Session)
        self.client = mock.Mock()
        self.session.create_client.return_value = self.client
        self.mock_config_writer = mock.Mock(spec=ConfigFileWriter)
        self.shared_config_file = 'shared-config-file'
        self.config_api = mock.Mock(spec=core.SharedConfigAPI)
        self.executor = core.Executor(
            step_handlers={
                'apicall': core.APICallExecutorStep(
                    core.APIInvoker(session=self.session),
                ),
                'sharedconfig': core.SharedConfigExecutorStep(
                    config_api=self.config_api,
                ),
                'define-variable': core.DefineVariableStep(),
                'merge-dict': core.MergeDictStep(),
                'load-data': core.LoadDataExecutorStep(),
                'dump-data': core.DumpDataExecutorStep(),
            }
        )

    def test_can_make_api_call_with_params(self):
        loaded = load_wizard("""
        execute:
          default:
            - type: apicall
              operation: iam.CreateUser
              params:
                UserName: admin
        """)
        self.executor.execute(loaded['execute'], {})
        self.session.create_client.assert_called_with('iam')
        self.client.create_user.assert_called_with(UserName='admin')

    def test_can_make_api_call_with_optional_params(self):
        loaded = load_wizard("""
        execute:
          default:
            - type: apicall
              operation: iam.CreateUser
              params:
                UserName: admin
              optional_params:
                Path: "/foo"
        """)
        self.executor.execute(loaded['execute'], {})
        self.session.create_client.assert_called_with('iam')
        self.client.create_user.assert_called_with(
            UserName='admin', Path='/foo'
        )

    def test_optional_params_not_passed_if_none(self):
        loaded = load_wizard("""
        execute:
          default:
            - type: apicall
              operation: iam.CreateUser
              params:
                UserName: admin
              optional_params:
                # Omitted because the value is null.
                Path: null
        """)
        self.executor.execute(loaded['execute'], {})
        self.session.create_client.assert_called_with('iam')
        self.client.create_user.assert_called_with(UserName='admin')

    def test_can_make_conditional_api_call(self):
        loaded = load_wizard("""
        execute:
          default:
            - type: apicall
              condition:
                variable: should_invoke
                equals: yes
              operation: iam.CreateUser
              params:
                UserName: admin
        """)
        self.executor.execute(loaded['execute'], {'should_invoke': 'no'})
        self.assertFalse(self.session.create_client.called)
        self.assertFalse(self.client.create_user.called)

    def test_can_make_conditional_on_env_var_not_exists(self):
        loaded = load_wizard("""
        execute:
          default:
            - type: apicall
              condition:
                variable: does_not_exist
                equals: null
              operation: iam.CreateUser
              params:
                UserName: admin
        """)
        self.executor.execute(loaded['execute'], {})
        self.assertTrue(self.session.create_client.called)
        self.assertTrue(self.client.create_user.called)

    def test_can_have_multiple_conditions(self):
        loaded = load_wizard("""
        execute:
          default:
            - type: apicall
              condition:
                - variable: foo
                  equals: one
                - variable: bar
                  equals: two
              operation: iam.CreateUser
              params:
                UserName: admin
        """)
        self.executor.execute(loaded['execute'], {'foo': 'one', 'bar': 'two'})
        self.assertTrue(self.session.create_client.called)
        self.assertTrue(self.client.create_user.called)

    def test_multiple_conditions_one_false(self):
        loaded = load_wizard("""
        execute:
          default:
            - type: apicall
              condition:
                - variable: foo
                  equals: one
                - variable: bar
                  equals: two
              operation: iam.CreateUser
              params:
                UserName: admin
        """)
        self.executor.execute(
            loaded['execute'], {'foo': 'one', 'bar': 'NOTTWO'}
        )
        self.assertFalse(self.session.create_client.called)

    def test_can_recursively_template_variables_in_params(self):
        loaded = load_wizard("""
        execute:
          default:
            - type: apicall
              operation: iam.CreateUser
              params:
                UserName: "{foo}"
                Nested:
                  Variable: "{foo}"
                ListType:
                  - one
                  - "{foo}"
                ComboNest:
                  Foo:
                    - Bar: "{foo}"
                    - Baz: "{foo}"
        """)
        self.executor.execute(loaded['execute'], {'foo': 'FOOVALUE'})
        self.session.create_client.assert_called_with('iam')
        expected_params = {
            'UserName': 'FOOVALUE',
            'Nested': {
                'Variable': 'FOOVALUE',
            },
            'ListType': ['one', 'FOOVALUE'],
            'ComboNest': {
                'Foo': [
                    {'Bar': 'FOOVALUE'},
                    {'Baz': 'FOOVALUE'},
                ]
            },
        }
        self.client.create_user.assert_called_with(**expected_params)

    def test_can_store_output_vars(self):
        loaded = load_wizard("""
        execute:
          default:
            - type: apicall
              operation: iam.CreateRole
              params:
                RoleName: admin
              output_var: role_arn
              query: Role.Arn
        """)
        params = {}
        self.client.create_role.return_value = {
            'Role': {'Arn': 'my-role-arn'},
        }
        self.executor.execute(loaded['execute'], params)
        self.client.create_role.assert_called_with(RoleName='admin')
        # We should have added 'role_arn' to the params dict and also
        # applied the jmespath query to the response before storing the
        # value.
        self.assertEqual(params['role_arn'], 'my-role-arn')

    def test_executes_multiple_groups(self):
        # We may introduce a 'next_step' similar to what you can do
        # in the planner, but for now, we just execute all steps sequentially
        # in the executor.
        loaded = load_wizard("""
        execute:
          default:
            - type: apicall
              operation: iam.CreateUser
              params:
                UserName: admin
          createrole:
            - type: apicall
              operation: iam.CreateRole
              params:
                RoleName: admin
        """)
        self.executor.execute(loaded['execute'], {})
        self.session.create_client.assert_called_with('iam')
        self.assertEqual(
            self.client.method_calls,
            [
                mock.call.create_user(UserName='admin'),
                mock.call.create_role(RoleName='admin'),
            ],
        )

    def test_can_write_to_config_file(self):
        loaded = load_wizard("""
        execute:
          default:
            - type: sharedconfig
              operation: SetValues
              profile: mydevprofile
              params:
                region: us-west-2
                output: json
        """)
        self.executor.execute(loaded['execute'], {})
        self.config_api.set_values.assert_called_with(
            {'region': 'us-west-2', 'output': 'json'},
            profile='mydevprofile',
        )

    def test_writes_to_default_profile_if_omitted(self):
        loaded = load_wizard("""
        execute:
          default:
            - type: sharedconfig
              operation: SetValues
              params:
                region: us-west-2
                output: json
        """)
        self.executor.execute(loaded['execute'], {})
        self.config_api.set_values.assert_called_with(
            {'region': 'us-west-2', 'output': 'json'}, profile=None
        )

    def test_can_expand_vars(self):
        loaded = load_wizard("""
        execute:
          default:
            - type: sharedconfig
              operation: SetValues
              params:
                region: "{foo}"
        """)
        variables = {'foo': 'bar'}
        self.executor.execute(loaded['execute'], variables)
        self.config_api.set_values.assert_called_with(
            {'region': 'bar'}, profile=None
        )

    def test_can_define_variables(self):
        loaded = load_wizard("""
        execute:
          default:
            - type: define-variable
              varname: myvar
              value:
                foo: bar
        """)
        variables = {}
        self.executor.execute(loaded['execute'], variables)
        self.assertEqual(variables, {'myvar': {'foo': 'bar'}})

    def test_can_reference_variables_in_definition(self):
        loaded = load_wizard("""
        execute:
          default:
            - type: define-variable
              varname: myvar
              value:
                foo: "{bar}"
        """)
        variables = {"bar": "value-of-bar"}
        self.executor.execute(loaded['execute'], variables)
        self.assertEqual(
            variables,
            {'myvar': {'foo': 'value-of-bar'}, 'bar': 'value-of-bar'},
        )

    def test_can_merge_dicts(self):
        loaded = load_wizard("""
        execute:
          default:
            - type: define-variable
              varname: myvar
              value:
                foo: original-foo
                bar: original-bar
                baz: original-baz
            - type: merge-dict
              output_var: result
              overlays:
              - "{myvar}"
              - bar: FIRST-NEW-BAR
                baz: FIRST-NEW-BAZ
              - baz: SECOND-NEW-BAZ
                newkey: newvalue
        """)
        variables = {}
        self.executor.execute(loaded['execute'], variables)
        expected = {
            'foo': 'original-foo',
            'bar': 'FIRST-NEW-BAR',
            'baz': 'SECOND-NEW-BAZ',
            'newkey': 'newvalue',
        }
        self.assertEqual(variables['result'], expected)

    def test_can_merge_nested_dicts(self):
        loaded = load_wizard("""
        execute:
          default:
            - type: define-variable
              varname: myvar
              value:
                foo:
                  bar:
                    baz: original-baz
                    baz2: original-baz2
                foo2: original-foo2
            - type: merge-dict
              output_var: result
              overlays:
              - "{myvar}"
              - foo:
                  bar:
                    baz: new-baz
                  bar2: new-bar2
        """)
        variables = {}
        self.executor.execute(loaded['execute'], variables)
        expected = {
            'foo': {
                'bar': {'baz': 'new-baz', 'baz2': 'original-baz2'},
                'bar2': 'new-bar2',
            },
            'foo2': 'original-foo2',
        }
        self.assertEqual(variables['result'], expected)

    def test_can_merge_dicts_with_vars(self):
        loaded = load_wizard("""
        execute:
          default:
            - type: define-variable
              varname: myvar
              value:
                foo:
                  bar:
                    baz: original-baz
                    baz2: original-baz2
                foo2: original-foo2
            - type: define-variable
              varname: myvar2
              value:
                foo:
                  bar:
                    baz: new-baz
                    baz2: original-baz2
                  bar2: new-bar2
            - type: merge-dict
              output_var: result
              overlays:
              - "{myvar}"
              - "{myvar2}"
        """)
        variables = {}
        self.executor.execute(loaded['execute'], variables)
        expected = {
            'foo': {
                'bar': {'baz': 'new-baz', 'baz2': 'original-baz2'},
                'bar2': 'new-bar2',
            },
            'foo2': 'original-foo2',
        }
        self.assertEqual(variables['result'], expected)

    def test_can_replace_merged_bar(self):
        loaded = load_wizard("""
        execute:
          default:
            - type: define-variable
              varname: myvar
              value:
                foo: bar
            - type: merge-dict
              output_var: myvar
              overlays:
              - "{myvar}"
              - foo: new
        """)
        variables = {}
        self.executor.execute(loaded['execute'], variables)
        self.assertEqual(variables, {'myvar': {'foo': 'new'}})

    def test_can_load_json(self):
        loaded = load_wizard("""
        execute:
          default:
            - type: load-data
              value: "{myvar}"
              output_var: loaded_myvar
              load_type: json
        """)
        variables = {'myvar': '{"foo": "bar"}'}
        self.executor.execute(loaded['execute'], variables)
        self.assertEqual(variables['loaded_myvar'], {'foo': 'bar'})

    def test_raises_error_for_unsupported_load_type(self):
        loaded = load_wizard("""
        execute:
          default:
            - type: load-data
              value: "{myvar}"
              output_var: loaded_myvar
              load_type: not-supported-type
        """)
        variables = {'myvar': '{"foo": "bar"}'}
        with self.assertRaisesRegex(ValueError, 'not-supported-type'):
            self.executor.execute(loaded['execute'], variables)

    def test_can_dump_json(self):
        loaded = load_wizard("""
        execute:
          default:
            - type: dump-data
              value: "{myvar}"
              output_var: dumped_myvar
              dump_type: json
        """)
        variables = {'myvar': {"foo": "bar"}}
        self.executor.execute(loaded['execute'], variables)
        self.assertEqual(variables['dumped_myvar'], '{"foo": "bar"}')

    def test_raises_error_for_unsupported_dump_type(self):
        loaded = load_wizard("""
        execute:
          default:
            - type: dump-data
              value: "{myvar}"
              output_var: dumped_myvar
              dump_type: not-supported-type
        """)
        variables = {'myvar': {"foo": "bar"}}
        with self.assertRaisesRegex(ValueError, 'not-supported-type'):
            self.executor.execute(loaded['execute'], variables)


class TestSharedConfigAPI(unittest.TestCase):
    def setUp(self):
        self.mock_session = mock.Mock(spec=Session)
        self.config_writer = mock.Mock(spec=ConfigFileWriter)
        self.config_filename = 'foo'
        self.mock_session.get_config_variable.return_value = (
            self.config_filename
        )
        self.config_api = core.SharedConfigAPI(
            self.mock_session, self.config_writer
        )

    def test_delegates_to_config_writer(self):
        self.config_api.set_values({'foo': 'bar'}, profile='bar')
        self.config_writer.update_config.assert_called_with(
            {'foo': 'bar', '__section__': 'profile bar'}, self.config_filename
        )

    def test_can_get_config_values(self):
        self.mock_session.get_config_variable.return_value = 'bar'
        self.assertEqual(self.config_api.get_value('foo'), 'bar')
        self.mock_session.get_config_variable.assert_called_with('foo')

    def test_can_list_profiles(self):
        self.mock_session.available_profiles = ['foo', 'bar']
        result = self.config_api.list_profiles()
        self.assertEqual(result, ['foo', 'bar'])


class TestRunner(unittest.TestCase):
    def setUp(self):
        self.planner = mock.Mock(spec=core.Planner)
        self.executor = mock.Mock(spec=core.Executor)
        self.runner = core.Runner(self.planner, self.executor)

    def test_can_delegate_to_obs(self):
        loaded = load_wizard("""
        plan:
          start:
            values:
              name:
                type: prompt
                description: Enter user name
        execute:
          default:
            - type: apicall
              operation: iam.CreateUser
              params:
                UserName: admin
        """)
        params = {'foo': 'bar'}
        self.planner.plan.return_value = params
        self.runner.run(wizard_spec=loaded)

        self.planner.plan.assert_called_with(loaded['plan'])
        self.executor.execute.assert_called_with(loaded['execute'], params)


class TestAPIInvoker(unittest.TestCase):
    def setUp(self):
        self.mock_session = mock.Mock(spec=Session)

    def get_call_args(self, session):
        return session.create_client.return_value.create_user.call_args

    def get_call_count(self, session):
        return session.create_client.return_value.create_user.call_count

    def get_paginate_call_args(self, session):
        client = session.create_client.return_value
        return client.get_paginator.return_value.paginate.call_args

    def test_can_make_api_call(self):
        invoker = core.APIInvoker(self.mock_session)
        invoker.invoke(
            'iam',
            'CreateUser',
            api_params={'UserName': 'admin'},
            plan_variables={},
        )
        call_method_args = self.get_call_args(self.mock_session)
        self.assertEqual(call_method_args, mock.call(UserName='admin'))

    def test_make_api_call_can_cache_response(self):
        invoker = core.APIInvoker(self.mock_session)
        invoker.invoke(
            'iam',
            'CreateUser',
            api_params={'UserName': 'admin'},
            plan_variables={},
            cache=True,
        )
        call_method_args = self.get_call_args(self.mock_session)
        self.assertEqual(call_method_args, mock.call(UserName='admin'))
        invoker.invoke(
            'iam',
            'CreateUser',
            api_params={'UserName': 'admin'},
            plan_variables={},
            cache=True,
        )
        self.assertEqual(self.get_call_count(self.mock_session), 1)

    def test_does_not_use_cache_when_different_input_parameters(self):
        invoker = core.APIInvoker(self.mock_session)
        invoker.invoke(
            'iam',
            'CreateUser',
            api_params={'UserName': 'admin'},
            plan_variables={},
            cache=True,
        )
        call_method_args = self.get_call_args(self.mock_session)
        self.assertEqual(call_method_args, mock.call(UserName='admin'))
        invoker.invoke(
            'iam',
            'CreateUser',
            api_params={'UserName': 'admin-different'},
            plan_variables={},
            cache=True,
        )
        call_method_args = self.get_call_args(self.mock_session)
        self.assertEqual(
            call_method_args, mock.call(UserName='admin-different')
        )
        self.assertEqual(self.get_call_count(self.mock_session), 2)

    def test_can_paginate(self):
        invoker = core.APIInvoker(self.mock_session)
        invoker.invoke(
            'iam',
            'ListPolicies',
            api_params={'Scope': 'All'},
            plan_variables={},
            paginate=True,
        )
        paginate_args = self.get_paginate_call_args(self.mock_session)
        self.assertEqual(paginate_args, mock.call(Scope='All'))

    def test_can_invoke_with_plan_variables(self):
        invoker = core.APIInvoker(self.mock_session)
        invoker.invoke(
            'iam',
            'CreateUser',
            api_params={'UserName': '{username}'},
            plan_variables={'username': 'admin'},
        )
        call_method_args = self.get_call_args(self.mock_session)
        self.assertEqual(call_method_args, mock.call(UserName='admin'))

    def test_can_open_file_with_builtin_function(self):
        invoker = core.APIInvoker(self.mock_session)
        with temporary_file('r+') as f:
            f.write('admin')
            f.flush()
            invoker.invoke(
                'iam',
                'CreateUser',
                # There's two parts to this test.  First, we have the
                # filename stored as a variable.
                plan_variables={'myfile': f.name},
                api_params={
                    # Then we reference the file with the builtin File
                    # type.  We should replace this entire value with
                    # the contents of the temp file ('admin').
                    'UserName': {'__wizard__:File': {'path': '{myfile}'}},
                },
            )
        call_method_args = self.get_call_args(self.mock_session)
        self.assertEqual(call_method_args, mock.call(UserName=b'admin'))


class TestTemplateStep(unittest.TestCase):
    def test_positive_condition_statement_for_equals(self):
        step_definition = {
            'type': 'template',
            'value': "{%if {allow} == True %}allow body{%   endif   %}",
        }
        parameters = {'allow': 'True'}
        step = core.TemplateStep()
        value = step.run_step(step_definition, parameters)
        self.assertEqual(value, 'allow body')

    def test_negative_condition_statement_for_equals(self):
        step_definition = {
            'type': 'template',
            'value': "{%if {allow} == True %}allow body{%   endif   %}",
        }
        parameters = {'allow': 'False'}
        step = core.TemplateStep()
        value = step.run_step(step_definition, parameters)
        self.assertEqual(value, '')

    def test_multiline_condition_statement(self):
        step_definition = {
            'type': 'template',
            'value': """{foo}
            {%if   {allow} == False    %}
not allow foo
            {% endif %}
   {%if   {allow} == True    %}
allow foo
        {% endif %}
more text""",
        }
        parameters = {
            'foo': 'foo parameter',
            'allow': 'True',
        }
        step = core.TemplateStep()
        value = step.run_step(step_definition, parameters)
        self.assertEqual(value, 'foo parameter\nallow foo\nmore text')

    def test_can_use_conditions_with_multiple_vars(self):
        step_definition = {
            'type': 'template',
            'value': "{%if {var1} == {var2} %}allow body{% endif %}",
        }
        parameters = {
            'var1': 'yes',
            'var2': 'yes',
        }
        step = core.TemplateStep()
        value = step.run_step(step_definition, parameters)
        self.assertEqual(value, 'allow body')

    def test_can_use_conditions_with_no_vars(self):
        step_definition = {
            'type': 'template',
            'value': "{%if yes == yes %}allow body{% endif %}",
        }
        parameters = {}
        step = core.TemplateStep()
        value = step.run_step(step_definition, parameters)
        self.assertEqual(value, 'allow body')

    def test_positive_condition_statements_with_not_equal(self):
        step_definition = {
            'type': 'template',
            'value': "{%if {first_var} != {second_var} %}not equals{% endif %}",
        }
        parameters = {
            'first_var': 'first_value',
            'second_var': 'second_value',
        }
        step = core.TemplateStep()
        value = step.run_step(step_definition, parameters)
        self.assertEqual(value, 'not equals')

    def test_negative_condition_statements_with_not_equal(self):
        step_definition = {
            'type': 'template',
            'value': "{%if {first_var} != {second_var} %}not equals{% endif %}",
        }
        parameters = {
            'first_var': 'same_value',
            'second_var': 'same_value',
        }
        step = core.TemplateStep()
        value = step.run_step(step_definition, parameters)
        self.assertEqual(value, '')

    def test_does_not_error_for_missing_vars_in_condition(self):
        step_definition = {
            'type': 'template',
            'value': "{%if {missing} == expected_value %}body{% endif %}",
        }
        parameters = {}
        step = core.TemplateStep()
        value = step.run_step(step_definition, parameters)
        self.assertEqual(value, '')

    def test_can_fetch_values(self):
        step_definition = {'type': 'template', 'value': "{foo} {bar}"}
        fake_wizard_values = FakeWizardValues()
        step = core.TemplateStep()
        value = step.run_step(step_definition, fake_wizard_values)
        self.assertEqual(fake_wizard_values, {'foo': 'foo', 'bar': 'bar'})
        self.assertEqual(value, 'foo bar')


class TestDumpDataStep:
    def test_dump_json_data(self):
        value = core.DumpDataStep().run_step(
            step_definition={
                'type': 'dump-data',
                'dump_type': 'json',
                'value': {"key": "value"},
            },
            parameters=FakeWizardValues(),
        )
        assert value == '{"key": "value"}'

    def test_dump_data_resolves_variables(self):
        wizard_values = FakeWizardValues()
        wizard_values.values['replace-me'] = {"replaced": "value"}
        value = core.DumpDataStep().run_step(
            step_definition={
                'type': 'dump-data',
                'dump_type': 'json',
                'value': '{replace-me}',
            },
            parameters=wizard_values,
        )
        assert value == '{"replaced": "value"}'

    def test_dump_unsupported_data_type(self):
        with pytest.raises(ValueError, match='not-supported-type'):
            core.DumpDataStep().run_step(
                step_definition={
                    'type': 'dump-data',
                    'dump_type': 'not-supported-type',
                    'value': {"key": "value"},
                },
                parameters=FakeWizardValues(),
            )


class TestLoadDataStep:
    def test_load_json_data(self):
        # Note double brackets are needed in order to escape the parameter
        # substitution as that is typically specified using brackets
        # (e.g. {<replace-param>})
        value = core.LoadDataStep().run_step(
            step_definition={
                'type': 'load-data',
                'load_type': 'json',
                'value': '{{"key": "value"}}',
            },
            parameters=FakeWizardValues(),
        )
        assert value == {"key": "value"}

    def test_load_data_resolves_variables(self):
        wizard_values = FakeWizardValues()
        wizard_values.values['replace-me'] = '{"replaced": "value"}'
        value = core.LoadDataStep().run_step(
            step_definition={
                'type': 'load-data',
                'load_type': 'json',
                'value': '{replace-me}',
            },
            parameters=wizard_values,
        )
        assert value == {"replaced": "value"}

    def test_load_unknown_data(self):
        with pytest.raises(ValueError, match='not-supported-type'):
            core.LoadDataStep().run_step(
                step_definition={
                    'type': 'load-data',
                    'load_type': 'not-supported-type',
                    'value': '{"key": "value"}',
                },
                parameters=FakeWizardValues(),
            )



# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import time
import uuid

from awscli.clidriver import AWSCLIEntryPoint
from awscli.customizations.sso.utils import AuthCodeFetcher, OpenBrowserHandler
from awscli.testutils import (
    BaseAWSCommandParamsTest,
    FileCreator,
    create_clidriver,
    mock,
)


class BaseSSOTest(BaseAWSCommandParamsTest):
    def setUp(self):
        super(BaseSSOTest, self).setUp()
        self.files = FileCreator()
        self.start_url = 'https://mysigin.com'
        self.sso_region = 'us-west-2'
        self.registration_scopes = None
        self.account = '012345678912'
        self.role_name = 'SSORole'
        self.config_file = self.files.full_path('config')
        self.environ['AWS_CONFIG_FILE'] = self.config_file
        self.set_config_file_content()
        self.access_token = 'foo.token.string'
        self.token_cache_dir = self.files.full_path('token-cache')
        self.token_cache_dir_patch = mock.patch(
            'awscli.customizations.sso.utils.SSO_TOKEN_DIR',
            self.token_cache_dir,
        )
        self.token_cache_dir_patch.start()
        self.open_browser_mock = mock.Mock(spec=OpenBrowserHandler)
        self.open_browser_patch = mock.patch(
            'awscli.customizations.sso.utils.OpenBrowserHandler',
            self.open_browser_mock,
        )
        self.open_browser_patch.start()

        self.fetcher_mock = mock.Mock(spec=AuthCodeFetcher)
        self.fetcher_mock.return_value.redirect_uri_without_port.return_value = 'http://127.0.0.1/oauth/callback'
        self.fetcher_mock.return_value.redirect_uri_with_port.return_value = (
            'http://127.0.0.1:55555/oauth/callback'
        )
        self.fetcher_mock.return_value.get_auth_code_and_state.return_value = (
            "abc",
            "00000000-0000-0000-0000-000000000000",
        )
        self.auth_code_fetcher_patch = mock.patch(
            'awscli.customizations.sso.utils.AuthCodeFetcher',
            self.fetcher_mock,
        )
        self.auth_code_fetcher_patch.start()

        self.uuid_mock = mock.Mock(
            return_value=uuid.UUID("00000000-0000-0000-0000-000000000000")
        )
        self.uuid_patch = mock.patch('uuid.uuid4', self.uuid_mock)
        self.uuid_patch.start()

        self.expires_in = 28800
        self.expiration_time = time.time() + 1000

    def tearDown(self):
        super(BaseSSOTest, self).tearDown()
        self.files.remove_all()
        self.open_browser_patch.stop()
        self.auth_code_fetcher_patch.stop()
        self.uuid_patch.stop()
        self.token_cache_dir_patch.stop()

    def add_oidc_device_responses(
        self, access_token, include_register_response=True
    ):
        responses = [
            # StartDeviceAuthorization response
            {
                'interval': 1,
                'expiresIn': 600,
                'userCode': 'foo',
                'deviceCode': 'foo-device-code',
                'verificationUri': 'https://sso.fake/device',
                'verificationUriComplete': 'https://sso.verify',
            },
            # CreateToken responses
            {
                'Error': {
                    'Code': 'AuthorizationPendingException',
                    'Message': 'Authorization is still pending',
                }
            },
            {
                'expiresIn': self.expires_in,
                'tokenType': 'Bearer',
                'accessToken': access_token,
            },
        ]
        if include_register_response:
            responses.insert(
                0,
                {
                    'clientSecretExpiresAt': self.expiration_time,
                    'clientId': 'device-client-id',
                    'clientSecret': 'device-client-secret',
                },
            )
        self.parsed_responses = responses

    def add_oidc_auth_code_responses(
        self, access_token, include_register_response=True
    ):
        responses = [
            # CreateToken responses
            {
                'expiresIn': self.expires_in,
                'tokenType': 'Bearer',
                'accessToken': access_token,
            }
        ]
        if include_register_response:
            responses.insert(
                0,
                {
                    'clientSecretExpiresAt': self.expiration_time,
                    'clientId': 'auth-client-id',
                    'clientSecret': 'auth-client-secret',
                },
            )
        self.parsed_responses = responses

    def assert_used_expected_sso_region(self, expected_region):
        self.assertIn(expected_region, self.last_request_dict['url'])

    def assert_device_browser_handler_called_with(
        self,
        userCode,
        verificationUri,
        verificationUriComplete,
    ):
        # assert_called_with is matching the __init__ parameters instead of
        # __call__, so verify the arguments we're interested in this way
        self.open_browser_mock.assert_called_once()
        _, kwargs = self.open_browser_mock.return_value.call_args
        self.assertEqual(userCode, kwargs['userCode'])
        self.assertEqual(verificationUri, kwargs['verificationUri'])
        self.assertEqual(
            verificationUriComplete, kwargs['verificationUriComplete']
        )

    def assert_auth_browser_handler_called_with(self, expected_scopes):
        # The endpoint is subject to the endpoint rules, and the
        # code_challenge is not fixed so assert against the rest of the url
        expected_url = (
            'authorize?'
            'response_type=code'
            '&client_id=auth-client-id'
            '&redirect_uri=http%3A%2F%2F127.0.0.1%3A55555%2Foauth%2Fcallback'
            '&state=00000000-0000-0000-0000-000000000000'
            '&code_challenge_method=S256'
            '&scopes=' + expected_scopes
        )

        self.open_browser_mock.assert_called_once()
        _, kwargs = self.open_browser_mock.return_value.call_args
        self.assertEqual(None, kwargs['userCode'])
        self.assertIn(expected_url, kwargs['verificationUri'])
        self.assertIn(expected_url, kwargs['verificationUriComplete'])

    def get_legacy_config(self):
        content = (
            f'[default]\n'
            f'sso_start_url={self.start_url}\n'
            f'sso_region={self.sso_region}\n'
            f'sso_role_name={self.role_name}\n'
            f'sso_account_id={self.account}\n'
        )
        return content

    def get_sso_session_config(self, session_name, include_profile=True):
        content = ''
        if include_profile:
            content += (
                f'[default]\n'
                f'sso_session={session_name}\n'
                f'sso_role_name={self.role_name}\n'
                f'sso_account_id={self.account}\n'
            )
        content += (
            f'[sso-session {session_name}]\n'
            f'sso_start_url={self.start_url}\n'
            f'sso_region={self.sso_region}\n'
        )
        if self.registration_scopes:
            scopes = ', '.join(self.registration_scopes)
            content += f'sso_registration_scopes={scopes}'
        return content

    def set_config_file_content(self, content=None):
        if content is None:
            content = self.get_legacy_config()
        self.files.create_file(self.config_file, content)
        # We need to recreate the driver (which includes its session) in order
        # for the config changes to be pulled in by the session.
        self.driver = create_clidriver()
        self.entry_point = AWSCLIEntryPoint(self.driver)



# Copyright (c) 2013 Amazon.com, Inc. or its affiliates.  All Rights Reserved
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish, dis-
# tribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the fol-
# lowing conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABIL-
# ITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT
# SHALL THE AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
#
import unittest

from awscli.table import MultiTable, Section, convert_to_vertical_table


class TestSection(unittest.TestCase):
    def setUp(self):
        self.section = Section()

    def test_add_row_tracks_max_widths(self):
        self.section.add_row(['one', 'two', 'three'])
        self.assertEqual(self.section.calculate_column_widths(), [3, 3, 5])
        self.section.add_row(['1234567', '1234567', '1234567'])
        self.assertEqual(self.section.calculate_column_widths(), [7, 7, 7])
        self.section.add_row(['a', 'a', 'a'])
        self.assertEqual(self.section.calculate_column_widths(), [7, 7, 7])
        self.section.add_row(['123456789', '1', '1'])
        self.assertEqual(self.section.calculate_column_widths(), [9, 7, 7])
        self.assertEqual(self.section.calculate_column_widths(1), [10, 8, 8])
        self.assertEqual(self.section.total_width(), 23)

    def test_add_row_also_tracks_header(self):
        self.section.add_header(['123456789', '12345', '1234567'])
        self.section.add_row(['a', 'a', 'a'])
        self.section.add_row(['aa', 'aa', 'aa'])
        self.section.add_row(['aaa', 'aaa', 'aaa'])
        self.assertEqual(self.section.calculate_column_widths(), [9, 5, 7])
        self.assertEqual(self.section.total_width(), 21)
        self.section.add_row(['aaa', '123456789', 'aaa'])
        self.assertEqual(self.section.calculate_column_widths(), [9, 9, 7])
        self.assertEqual(
            self.section.total_width(padding=3, with_border=True), 36
        )

    def test_max_width_with_scaling_perfect_scaling(self):
        self.section.add_row(['one', 'two', 'three'])
        self.section.add_row(['1234567', '1234567', '1234567'])
        # Perfect scaling, exactly double.
        widths = self.section.calculate_column_widths(max_width=42)
        self.assertEqual(widths, [14, 14, 14])

    def test_max_width_scaling_one_unit_short(self):
        self.section.add_row(['one', 'two', 'three'])
        self.section.add_row(['1234567', '1234567', '1234567'])
        widths = self.section.calculate_column_widths(max_width=41)
        self.assertEqual(widths, [13, 14, 14])

    def test_width_with_full_width_characters(self):
        self.section.add_row(['\u4e00', 'one'])
        self.assertEqual(self.section.calculate_column_widths(), [2, 3])

    def test_max_width_scaling_is_negative(self):
        self.section.add_row(['12345', '12345'])
        widths = self.section.calculate_column_widths(max_width=17)
        self.assertEqual(widths, [8, 9])

    def test_allow_sections_to_be_padded(self):
        self.section.add_row(['one', 'two', 'three'])
        self.section.add_row(['1234567', '1234567', '1234567'])
        self.assertEqual(self.section.total_width(padding=2), 27)
        self.assertEqual(
            self.section.total_width(padding=2, outer_padding=1), 29
        )

    def test_title_accounts_for_outer_padding(self):
        self.section.add_row(['a', 'b', 'c'])
        self.section.add_title('123456789')
        self.assertEqual(
            self.section.total_width(padding=2, outer_padding=3), 17
        )

    def test_unicode_text_row(self):
        self.section.add_row([1])
        self.section.add_row(['check'])
        self.section.add_row(['\u2713'])
        self.assertEqual(self.section.rows, [['1'], ['check'], ['\u2713']])


class TestMultiTable(unittest.TestCase):
    def setUp(self):
        self.table = MultiTable()

    def test_max_width_calculation(self):
        self.table.add_title('foo')
        self.table.add_row_header(['one', 'two', 'three'])
        self.table.add_row(['one', 'two', 'three'])
        self.table.new_section('bar')
        self.table.add_row_header(['one', 'two'])
        self.table.add_row(['12345', '1234567'])


class TestVerticalTableConversion(unittest.TestCase):
    def setUp(self):
        self.table = MultiTable()

    def test_convert_section_to_vertical(self):
        self.table.add_title('foo')
        self.table.add_row_header(['key1', 'key2', 'key3'])
        self.table.add_row(['val1', 'val2', 'val3'])
        convert_to_vertical_table(self.table._sections)
        # To convert to a vertical table, there should be no headers:
        section = self.table._sections[0]
        self.assertEqual(section.headers, [])
        # Then we should create a two column row with key val pairs.
        self.assertEqual(
            section.rows,
            [['key1', 'val1'], ['key2', 'val2'], ['key3', 'val3']],
        )


if __name__ == '__main__':
    unittest.main()



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.compat import compat_open
from awscli.testutils import BaseAWSCommandParamsTest, temporary_file


class TestRunInstances(BaseAWSCommandParamsTest):
    prefix = 'ec2 run-instances'

    def assert_run_instances_call(self, args, result):
        if not isinstance(args, list):
            args_list = (self.prefix + args).split()
        else:
            args_list = self.prefix.split() + args
        self.assert_params_for_cmd(
            args_list, result, ignore_params=['ClientToken']
        )

    def test_no_count(self):
        args = ' --image-id ami-foobar'
        result = {'ImageId': 'ami-foobar', 'MaxCount': 1, 'MinCount': 1}
        self.assert_run_instances_call(args, result)

    def test_count_scalar(self):
        args = ' --image-id ami-foobar --count 2'
        result = {'ImageId': 'ami-foobar', 'MaxCount': 2, 'MinCount': 2}
        self.assert_run_instances_call(args, result)

    def test_user_data(self):
        data = '\u0039'
        with temporary_file('r+') as tmp:
            with compat_open(tmp.name, 'w') as f:
                f.write(data)
                f.flush()
                args = ' --image-id foo --user-data file://%s' % f.name
                result = {
                    'ImageId': 'foo',
                    'MaxCount': 1,
                    'MinCount': 1,
                    # base64 encoded content of utf-8 encoding of data.
                    'UserData': 'OQ==',
                }
            self.assert_run_instances_call(args, result)

    def test_count_range(self):
        args = ' --image-id ami-foobar --count 5:10'
        result = {'ImageId': 'ami-foobar', 'MaxCount': 10, 'MinCount': 5}
        self.assert_run_instances_call(args, result)

    def test_count_in_json_only(self):
        input_json = '{"ImageId":"ami-xxxx","MaxCount":9,"MinCount":5}'
        args = ' --cli-input-json ' + input_json
        result = {'ImageId': 'ami-xxxx', 'MaxCount': 9, 'MinCount': 5}
        self.assert_run_instances_call(args, result)

    def test_count_in_cli_and_in_json(self):
        input_json = '{"ImageId":"ami-xxxx","MaxCount":9,"MinCount":5}'
        args = ' --count 3 --cli-input-json ' + input_json
        result = {'ImageId': 'ami-xxxx', 'MaxCount': 3, 'MinCount': 3}
        self.assert_run_instances_call(args, result)

    def test_block_device_mapping(self):
        args_list = ' --image-id ami-foobar --count 1'.split()
        # We're switching to list form because we need to test
        # when there's leading spaces.  This is the CLI equivalent
        # of --block-dev-mapping ' [{"device_name" ...'
        # (note the space between ``'`` and ``[``)
        args_list.append('--block-device-mapping')
        args_list.append(
            ' [{"DeviceName":"/dev/sda1","Ebs":{"VolumeSize":20}}]'
        )
        result = {
            'BlockDeviceMappings': [
                {'DeviceName': '/dev/sda1', 'Ebs': {'VolumeSize': 20}},
            ],
            'ImageId': 'ami-foobar',
            'MaxCount': 1,
            'MinCount': 1,
        }
        self.assert_run_instances_call(args_list, result)

    def test_secondary_ip_address(self):
        args = ' --image-id ami-foobar --count 1 '
        args += '--secondary-private-ip-addresses 10.0.2.106'
        args_list = (self.prefix + args).split()
        result = {
            'ImageId': 'ami-foobar',
            'NetworkInterfaces': [
                {
                    'DeviceIndex': 0,
                    'PrivateIpAddresses': [
                        {'Primary': False, 'PrivateIpAddress': '10.0.2.106'}
                    ],
                }
            ],
            'MaxCount': 1,
            'MinCount': 1,
        }
        self.assert_run_instances_call(args, result)

    def test_secondary_ip_address_with_subnet(self):
        args = ' --image-id ami-foobar --count 1 --subnet subnet-12345678 '
        args += '--secondary-private-ip-addresses 10.0.2.106'
        result = {
            'ImageId': 'ami-foobar',
            'NetworkInterfaces': [
                {
                    'DeviceIndex': 0,
                    'SubnetId': 'subnet-12345678',
                    'PrivateIpAddresses': [
                        {'Primary': False, 'PrivateIpAddress': '10.0.2.106'}
                    ],
                }
            ],
            'MaxCount': 1,
            'MinCount': 1,
        }
        self.assert_run_instances_call(args, result)

    def test_secondary_ip_addresses(self):
        args = ' --image-id ami-foobar --count 1 '
        args += '--secondary-private-ip-addresses 10.0.2.106 10.0.2.107'
        result = {
            'ImageId': 'ami-foobar',
            'NetworkInterfaces': [
                {
                    'DeviceIndex': 0,
                    'PrivateIpAddresses': [
                        {'Primary': False, 'PrivateIpAddress': '10.0.2.106'},
                        {'Primary': False, 'PrivateIpAddress': '10.0.2.107'},
                    ],
                }
            ],
            'MaxCount': 1,
            'MinCount': 1,
        }
        self.assert_run_instances_call(args, result)

    def test_secondary_ip_address_count(self):
        args = ' --image-id ami-foobar --count 1 '
        args += '--secondary-private-ip-address-count 4'
        result = {
            'NetworkInterfaces': [
                {'DeviceIndex': 0, 'SecondaryPrivateIpAddressCount': 4}
            ],
            'ImageId': 'ami-foobar',
            'MaxCount': 1,
            'MinCount': 1,
        }
        self.assert_run_instances_call(args, result)

    def test_secondary_ip_address_count_with_subnet(self):
        args = ' --image-id ami-foobar --count 1 --subnet subnet-12345678 '
        args += '--secondary-private-ip-address-count 4'
        result = {
            'NetworkInterfaces': [
                {
                    'DeviceIndex': 0,
                    'SubnetId': 'subnet-12345678',
                    'SecondaryPrivateIpAddressCount': 4,
                }
            ],
            'ImageId': 'ami-foobar',
            'MaxCount': 1,
            'MinCount': 1,
        }
        self.assert_run_instances_call(args, result)

    def test_associate_public_ip_address(self):
        args = ' --image-id ami-foobar --count 1 --subnet-id subnet-12345678 '
        args += '--associate-public-ip-address'
        result = {
            'NetworkInterfaces': [
                {
                    'DeviceIndex': 0,
                    'AssociatePublicIpAddress': True,
                    'SubnetId': 'subnet-12345678',
                },
            ],
            'ImageId': 'ami-foobar',
            'MaxCount': 1,
            'MinCount': 1,
        }
        self.assert_run_instances_call(args, result)

    def test_associate_public_ip_address_switch_order(self):
        args = ' --image-id ami-foobar --count 1 '
        args += '--associate-public-ip-address --subnet-id subnet-12345678'
        result = {
            'NetworkInterfaces': [
                {
                    'DeviceIndex': 0,
                    'AssociatePublicIpAddress': True,
                    'SubnetId': 'subnet-12345678',
                }
            ],
            'ImageId': 'ami-foobar',
            'MaxCount': 1,
            'MinCount': 1,
        }
        self.assert_run_instances_call(args, result)

    def test_no_associate_public_ip_address(self):
        args = ' --image-id ami-foobar --count 1  --subnet-id subnet-12345678 '
        args += '--no-associate-public-ip-address'
        result = {
            'ImageId': 'ami-foobar',
            'NetworkInterfaces': [
                {
                    'AssociatePublicIpAddress': False,
                    'DeviceIndex': 0,
                    'SubnetId': 'subnet-12345678',
                }
            ],
            'MaxCount': 1,
            'MinCount': 1,
        }
        self.assert_run_instances_call(args, result)

    def test_subnet_alone(self):
        args = ' --image-id ami-foobar --count 1 --subnet-id subnet-12345678'
        result = {
            'SubnetId': 'subnet-12345678',
            'ImageId': 'ami-foobar',
            'MaxCount': 1,
            'MinCount': 1,
        }
        self.assert_run_instances_call(args, result)

    def test_associate_public_ip_address_and_group_id(self):
        args = ' --image-id ami-foobar --count 1 '
        args += '--security-group-id sg-12345678 '
        args += '--associate-public-ip-address --subnet-id subnet-12345678'
        result = {
            'NetworkInterfaces': [
                {
                    'DeviceIndex': 0,
                    'AssociatePublicIpAddress': True,
                    'SubnetId': 'subnet-12345678',
                    'Groups': ['sg-12345678'],
                }
            ],
            'ImageId': 'ami-foobar',
            'MaxCount': 1,
            'MinCount': 1,
        }
        self.assert_run_instances_call(args, result)

    def test_group_id_alone(self):
        args = ' --image-id ami-foobar --count 1 '
        args += '--security-group-id sg-12345678'
        result = {
            'SecurityGroupIds': ['sg-12345678'],
            'ImageId': 'ami-foobar',
            'MaxCount': 1,
            'MinCount': 1,
        }
        self.assert_run_instances_call(args, result)

    def test_associate_public_ip_address_and_private_ip_address(self):
        args = ' --image-id ami-foobar --count 1 '
        args += '--private-ip-address 10.0.0.200 '
        args += '--associate-public-ip-address --subnet-id subnet-12345678'
        result = {
            'NetworkInterfaces': [
                {
                    'DeviceIndex': 0,
                    'AssociatePublicIpAddress': True,
                    'SubnetId': 'subnet-12345678',
                    'PrivateIpAddresses': [
                        {'PrivateIpAddress': '10.0.0.200', 'Primary': True}
                    ],
                }
            ],
            'ImageId': 'ami-foobar',
            'MaxCount': 1,
            'MinCount': 1,
        }
        self.assert_run_instances_call(args, result)

    def test_private_ip_address_alone(self):
        args = ' --image-id ami-foobar --count 1 '
        args += '--private-ip-address 10.0.0.200'
        result = {
            'PrivateIpAddress': '10.0.0.200',
            'ImageId': 'ami-foobar',
            'MaxCount': 1,
            'MinCount': 1,
        }
        self.assert_run_instances_call(args, result)

    def test_ipv6_address_count_and_associate_public_ip_address(self):
        args = ' --associate-public-ip-address'
        args += ' --ipv6-address-count 5 --image-id ami-foobar --count 1'
        expected = {
            'NetworkInterfaces': [
                {
                    'DeviceIndex': 0,
                    'AssociatePublicIpAddress': True,
                    'Ipv6AddressCount': 5,
                }
            ],
            'ImageId': 'ami-foobar',
            'MaxCount': 1,
            'MinCount': 1,
        }
        self.assert_run_instances_call(args, expected)

    def test_ipv6_addresses_and_associate_public_ip_address(self):
        args = ' --associate-public-ip-address --count 1'
        args += ' --ipv6-addresses Ipv6Address=::1 --image-id ami-foobar '
        expected = {
            'NetworkInterfaces': [
                {
                    'DeviceIndex': 0,
                    'AssociatePublicIpAddress': True,
                    'Ipv6Addresses': [{'Ipv6Address': '::1'}],
                }
            ],
            'ImageId': 'ami-foobar',
            'MaxCount': 1,
            'MinCount': 1,
        }
        self.assert_run_instances_call(args, expected)

    def test_enable_primary_ipv6_and_associate_public_ip_address(self):
        args = ' --associate-public-ip-address'
        args += ' --enable-primary-ipv6 --image-id ami-foobar --count 1'
        expected = {
            'NetworkInterfaces': [
                {
                    'DeviceIndex': 0,
                    'AssociatePublicIpAddress': True,
                    'PrimaryIpv6': True,
                }
            ],
            'ImageId': 'ami-foobar',
            'MaxCount': 1,
            'MinCount': 1,
        }
        self.assert_run_instances_call(args, expected)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import botocore.client

from awscli.autocomplete import parser
from awscli.autocomplete.completer import CompletionResult
from awscli.autocomplete.serverside.servercomp import (
    LazyClientCreator,
    ServerSideCompleter,
)
from awscli.testutils import mock, unittest
from tests.unit.autocomplete import InMemoryIndex


class FakeCompletionLookup:
    def __init__(self, completion_data):
        self.completion_data = completion_data

    def get_server_completion_data(self, lineage, command_name, param_name):
        key = (tuple(lineage), command_name, param_name)
        try:
            return self.completion_data[key]
        except KeyError:
            return None


class TestServerSideAutocompleter(unittest.TestCase):
    def setUp(self):
        self.index = InMemoryIndex(
            {
                'command_names': {
                    '': [('aws', None)],
                    'aws': [('ec2', None), ('iam', None), ('s3', None)],
                    'aws.iam': [
                        ('delete-user-policy', None),
                        ('delete-user', None),
                    ],
                },
                'arg_names': {
                    '': {
                        'aws': ['region', 'endpoint-url', 'profile'],
                    },
                    'aws.iam': {
                        'delete-user-policy': ['policy-name'],
                        'delete-user': ['user-name'],
                    },
                },
                'arg_data': {
                    '': {
                        'aws': {
                            'endpoint-url': (
                                'endpoint-url',
                                'string',
                                'aws',
                                '',
                                None,
                                False,
                                False,
                            ),
                            'region': (
                                'region',
                                'string',
                                'aws',
                                '',
                                None,
                                False,
                                False,
                            ),
                            'profile': (
                                'profile',
                                'string',
                                'aws',
                                '',
                                None,
                                False,
                                False,
                            ),
                        }
                    },
                    'aws.iam': {
                        'delete-user-policy': {
                            'policy-name': (
                                'policy-name',
                                'string',
                                'delete-user-policy',
                                'aws.iam.',
                                None,
                                False,
                                False,
                            ),
                        },
                        'delete-user': {
                            'user-name': (
                                'user-name',
                                'string',
                                'delete-user',
                                'aws.iam.',
                                None,
                                False,
                                False,
                            ),
                        },
                    },
                },
            }
        )
        key = (('aws', 'iam'), 'delete-user-policy', 'policy-name')
        self.completion_data = {
            key: {
                'completions': [
                    {
                        "parameters": {},
                        "service": "iam",
                        "operation": "list_policies",
                        "jp_expr": "Policies[].PolicyName",
                    }
                ]
            }
        }
        self.parser = parser.CLIParser(self.index)
        self.mock_client = mock.Mock()
        self.mock_create_client = mock.Mock()
        self.mock_create_client.create_client.return_value = self.mock_client
        self.completion_lookup = FakeCompletionLookup(self.completion_data)
        self.completer = ServerSideCompleter(
            self.completion_lookup, self.mock_create_client
        )

    def test_does_not_complete_if_unparsed_items(self):
        parsed = self.parser.parse('aws foo ')
        self.assertIsNone(self.completer.complete(parsed))

    def test_does_complete_if_current_fragment_is_none(self):
        parsed = self.parser.parse('aws')
        self.assertIsNone(self.completer.complete(parsed))

    def test_does_not_complete_if_not_on_param_value(self):
        parsed = self.parser.parse('aws iam delete-user-policy --poli')
        self.assertIsNone(self.completer.complete(parsed))

    def test_no_completion_data_returns_none(self):
        parsed = self.parser.parse('aws iam delete-user --user-name ')
        self.assertIsNone(self.completer.complete(parsed))

    def test_can_prefix_match_results(self):
        self.mock_client.list_policies.return_value = {
            'Policies': [
                {'PolicyName': 'a1'},
                {'PolicyName': 'b1'},
                {'PolicyName': 'a2'},
                {'PolicyName': 'b2'},
            ],
        }
        parsed = self.parser.parse(
            'aws iam delete-user-policy --policy-name a'
        )
        results = self.completer.complete(parsed)
        self.assertEqual(
            results, [CompletionResult('a1', -1), CompletionResult('a2', -1)]
        )

    def test_returns_all_results_when_current_fragment_empty(self):
        self.mock_client.list_policies.return_value = {
            'Policies': [
                {'PolicyName': 'a1'},
                {'PolicyName': 'b1'},
                {'PolicyName': 'a2'},
                {'PolicyName': 'b2'},
            ],
        }
        parsed = self.parser.parse('aws iam delete-user-policy --policy-name ')
        results = self.completer.complete(parsed)
        self.assertEqual(
            results,
            [
                CompletionResult('a1', 0),
                CompletionResult('b1', 0),
                CompletionResult('a2', 0),
                CompletionResult('b2', 0),
            ],
        )

    def test_returns_empty_list_on_client_error(self):
        self.mock_client.list_policies.side_effect = Exception(
            "Something went wrong."
        )
        parsed = self.parser.parse('aws iam delete-user-policy --policy-name ')
        results = self.completer.complete(parsed)
        self.assertEqual(results, [])

    def test_returns_empty_list_when_jmespath_doesnt_match(self):
        self.mock_client.list_policies.return_value = {
            'WrongKeySomehow': [],
        }
        parsed = self.parser.parse('aws iam delete-user-policy --policy-name ')
        results = self.completer.complete(parsed)
        self.assertEqual(results, [])

    def test_region_and_profile_passed_to_create_client(self):
        parsed = self.parser.parse(
            'aws iam delete-user-policy '
            '--region us-west-2 --profile profile1 '
            '--policy-name '
        )
        self.completer.complete(parsed)
        self.mock_create_client.create_client.assert_called_with(
            'iam', parsed_profile='profile1', parsed_region='us-west-2'
        )


class TestLazyClientCreator(unittest.TestCase):
    def test_can_create_client(self):
        creator = LazyClientCreator()
        client = creator.create_client(
            'iam',
            aws_access_key_id='foo',
            aws_secret_access_key='bar',
            region_name='us-west-2',
        )
        self.assertIsInstance(client, botocore.client.BaseClient)
        # Sanity check it's an IAM client.
        self.assertTrue(hasattr(client, 'list_users'))
        # Verify we can create another client.
        new_client = creator.create_client(
            'iam',
            aws_access_key_id='foo',
            aws_secret_access_key='bar',
            region_name='us-west-2',
        )
        self.assertIsInstance(new_client, botocore.client.BaseClient)

    def test_can_use_session_cache(self):
        creator = LazyClientCreator()
        creator.create_session = mock.Mock()
        creator.create_client('iam', parsed_profile='foo')
        creator.create_client('rds', parsed_profile='foo')
        self.assertEqual(creator.create_session.call_count, 1)

    def test_not_use_session_cache_for_different_profiles(self):
        creator = LazyClientCreator()
        creator.create_session = mock.Mock()
        creator.create_client('iam', parsed_profile='foo')
        creator.create_client('rds', parsed_profile='bar')
        self.assertEqual(creator.create_session.call_count, 2)



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.customizations.cloudsearchdomain import validate_endpoint_url
from awscli.customizations.exceptions import ParamValidationError
from awscli.help import PagingHelpRenderer
from awscli.testutils import BaseAWSCommandParamsTest, mock, unittest


class TestSearchCommand(BaseAWSCommandParamsTest):
    prefix = 'cloudsearchdomain search '

    def test_search_with_query(self):
        cmd = self.prefix.split()
        cmd += [
            '--endpoint-url',
            'http://example.com/',
            # Note we're also verifying that --query is renamed to
            # --search-query from argrename.py.
            '--search-query',
            'George Lucas',
            '--query-options',
            '{"defaultOperator":"and","fields":["directors^10"]}',
        ]

        expected = {
            'query': 'George Lucas',
            'queryOptions': '{"defaultOperator":"and","fields":["directors^10"]}',
        }
        self.assert_params_for_cmd(cmd, expected)

    def test_endpoint_is_required(self):
        cmd = self.prefix.split()
        cmd += ['--search-query', 'foo']
        stderr = self.run_cmd(cmd, expected_rc=252)[1]
        self.assertIn('--endpoint-url is required', stderr)

    def test_endpoint_not_required_for_help(self):
        cmd = self.prefix + 'help'
        with mock.patch('awscli.help.get_renderer') as get_renderer:
            mock_render = mock.Mock(spec=PagingHelpRenderer)
            get_renderer.return_value = mock_render
            stdout, stderr, rc = self.run_cmd(cmd, expected_rc=None)
            # If we get this far we've succeeded, but we can do
            # a quick sanity check and make sure the service name is
            # in the stdout help text.
            self.assertIn(stdout, 'cloudsearchdomain')


class TestCloudsearchDomainHandler(unittest.TestCase):
    def test_validate_endpoint_url_is_none(self):
        parsed_globals = mock.Mock()
        parsed_globals.endpoint_url = None
        # Method should return instantiated exception.
        err = validate_endpoint_url(parsed_globals)
        self.assertTrue(isinstance(err, ParamValidationError))


if __name__ == "__main__":
    unittest.main()



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.autocomplete import main


def test_smoke_test_completer():
    # Verify we can use the completer using the default auto-complete index
    # that should have been built and installed as part of building the
    # CLI distribution.
    completions = _autocomplete('aws ec2 desc')
    # The API can change so we won't assert a specific list, but we'll
    # pick a few operations that we know will always be there.
    completion_strings = [c.name for c in completions]
    assert 'describe-instances' in completion_strings
    assert 'describe-regions' in completion_strings

    completions = _autocomplete('aws dynamodb describe-tab')
    completion_strings = [c.name for c in completions]
    assert all(
        completion.startswith('describe-table')
        for completion in completion_strings
    )


def _autocomplete(command_line):
    completer = main.create_autocompleter()
    return completer.autocomplete(command_line)



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from botocore.model import DenormalizedStructureBuilder

from awscli.compat import StringIO
from awscli.customizations.generatecliskeleton import (
    GenerateCliSkeletonArgument,
)
from awscli.testutils import capture_output, mock, unittest


class TestGenerateCliSkeleton(unittest.TestCase):
    def setUp(self):
        self.session = mock.Mock()
        # Create a mock service operation object
        self.service_operation = mock.Mock()

        # Make an arbitrary input model shape.
        self.input_shape = {
            'A': {
                'type': 'structure',
                'members': {
                    'B': {'type': 'string'},
                },
            }
        }
        shape = (
            DenormalizedStructureBuilder()
            .with_members(self.input_shape)
            .build_model()
        )
        self.operation_model = mock.Mock(input_shape=shape)
        self.argument = GenerateCliSkeletonArgument(
            self.session, self.operation_model
        )

        # This is what the json should should look like after being
        # generated to standard output.
        self.ref_json_output = '{\n    "A": {\n        "B": ""\n    }\n}\n'

    def assert_skeleton_equals(
        self, arg_value, expected_output, expected_rc=0, input_shape=None
    ):
        argument = self.argument
        if input_shape is not None:
            argument = self.create_argument(input_shape)
        parsed_args = mock.Mock(generate_cli_skeleton=arg_value)
        with capture_output() as output:
            rc = argument.generate_skeleton(
                service_operation=self.service_operation,
                parsed_args=parsed_args,
                parsed_globals=None,
                call_parameters=None,
            )
            self.assertEqual(rc, expected_rc)
            self.assertEqual(output.stdout.getvalue(), expected_output)

    def create_argument(self, input_shape):
        builder = DenormalizedStructureBuilder().with_members(input_shape)
        operation_model = mock.Mock(input_shape=builder.build_model())
        return GenerateCliSkeletonArgument(self.session, operation_model)

    def test_register_argument_action(self):
        register_args = self.session.register.call_args_list
        self.assertEqual(register_args[0][0][0], 'calling-command.*')
        self.assertEqual(
            register_args[0][0][1], self.argument.generate_skeleton
        )

    def test_no_override_required_args_when_output(self):
        argument_table = {}
        mock_arg = mock.Mock()
        mock_arg.required = True
        argument_table['required-arg'] = mock_arg
        args = ['--generate-cli-skeleton', 'output']
        self.argument.override_required_args(argument_table, args)
        self.assertTrue(argument_table['required-arg'].required)

    def test_override_required_args_when_input(self):
        argument_table = {}
        mock_arg = mock.Mock()
        mock_arg.required = True
        argument_table['required-arg'] = mock_arg
        args = ['--generate-cli-skeleton']
        self.argument.override_required_args(argument_table, args)
        self.assertFalse(argument_table['required-arg'].required)

    def test_override_required_args_when_output_present_but_not_value(self):
        argument_table = {}
        mock_arg = mock.Mock()
        mock_arg.required = True
        argument_table['required-arg'] = mock_arg
        args = ['--generate-cli-skeleton', '--some-other-param', 'output']
        self.argument.override_required_args(argument_table, args)
        self.assertFalse(argument_table['required-arg'].required)

    def test_generate_json_skeleton(self):
        parsed_args = mock.Mock()
        parsed_args.generate_cli_skeleton = 'input'
        with mock.patch('sys.stdout', StringIO()) as mock_stdout:
            rc = self.argument.generate_skeleton(
                service_operation=self.service_operation,
                call_parameters=None,
                parsed_args=parsed_args,
                parsed_globals=None,
            )
            # Ensure the contents printed to standard output are correct.
            self.assertEqual(self.ref_json_output, mock_stdout.getvalue())
            # Ensure it is the correct return code of zero.
            self.assertEqual(rc, 0)

    def test_no_generate_json_skeleton(self):
        parsed_args = mock.Mock()
        parsed_args.generate_cli_skeleton = None
        with mock.patch('sys.stdout', StringIO()) as mock_stdout:
            rc = self.argument.generate_skeleton(
                service_operation=self.service_operation,
                call_parameters=None,
                parsed_args=parsed_args,
                parsed_globals=None,
            )
            # Ensure nothing is printed to standard output
            self.assertEqual('', mock_stdout.getvalue())
            # Ensure nothing is returned because it was never called.
            self.assertEqual(rc, None)

    def test_generate_json_skeleton_no_input_shape(self):
        parsed_args = mock.Mock()
        parsed_args.generate_cli_skeleton = 'input'
        # Set the input shape to ``None``.
        self.argument = GenerateCliSkeletonArgument(
            self.session, mock.Mock(input_shape=None)
        )
        with mock.patch('sys.stdout', StringIO()) as mock_stdout:
            rc = self.argument.generate_skeleton(
                service_operation=self.service_operation,
                call_parameters=None,
                parsed_args=parsed_args,
                parsed_globals=None,
            )
            # Ensure the contents printed to standard output are correct,
            # which should be an empty dictionary.
            self.assertEqual('{}\n', mock_stdout.getvalue())
            # Ensure it is the correct return code of zero.
            self.assertEqual(rc, 0)

    def test_generate_json_skeleton_with_timestamp(self):
        parsed_args = mock.Mock()
        parsed_args.generate_cli_skeleton = 'input'
        input_shape = {
            'A': {
                'type': 'structure',
                'members': {
                    'B': {'type': 'timestamp'},
                },
            }
        }
        shape = (
            DenormalizedStructureBuilder()
            .with_members(input_shape)
            .build_model()
        )
        operation_model = mock.Mock(input_shape=shape)
        argument = GenerateCliSkeletonArgument(self.session, operation_model)
        with mock.patch('sys.stdout', StringIO()) as mock_stdout:
            rc = argument.generate_skeleton(
                call_parameters=None,
                parsed_args=parsed_args,
                parsed_globals=None,
            )
            self.assertEqual(
                '{\n'
                '    "A": {\n'
                '        "B": "1970-01-01T00:00:00"\n'
                '    }\n'
                '}\n',
                mock_stdout.getvalue(),
            )
            self.assertEqual(rc, 0)

    def test_generate_yaml_input_skeleton(self):
        expected = "A:\n" "  B: ''\n"
        self.assert_skeleton_equals('yaml-input', expected)

    def test_generate_yaml_input_with_bytes(self):
        input_shape = {
            'A': {
                'type': 'structure',
                'members': {
                    'B': {'type': 'blob'},
                },
            }
        }
        expected = "A:\n" "  B: !!binary ''\n"
        self.assert_skeleton_equals(
            'yaml-input', expected, input_shape=input_shape
        )

    def test_generate_yaml_input_with_comments(self):
        input_shape = {'A': {'type': 'string', 'documentation': 'docstring'}}
        expected = "A: ''  # docstring.\n"
        self.assert_skeleton_equals(
            'yaml-input', expected, input_shape=input_shape
        )

    def test_generate_yaml_input_with_nested_comments(self):
        input_shape = {
            'A': {
                'type': 'structure',
                'documentation': 'top-level docstring',
                'members': {
                    'B': {
                        'type': 'string',
                        'documentation': 'nested docstring',
                    }
                },
            }
        }
        expected = (
            "A:  # top-level docstring.\n" "  B: ''  # nested docstring.\n"
        )
        self.assert_skeleton_equals(
            'yaml-input', expected, input_shape=input_shape
        )

    def test_generate_yaml_input_with_required_member(self):
        input_shape = {
            'A': {
                'type': 'structure',
                'documentation': 'top-level docstring',
                'members': {
                    'B': {
                        'type': 'string',
                        'documentation': 'nested docstring',
                    }
                },
                'required': ['B'],
            }
        }
        expected = (
            "A:  # top-level docstring.\n"
            "  B: ''  # [REQUIRED] nested docstring.\n"
        )
        self.assert_skeleton_equals(
            'yaml-input', expected, input_shape=input_shape
        )

    def test_generate_yaml_input_ignores_newline(self):
        input_shape = {
            'A': {'type': 'string', 'documentation': 'one-line\ntwo-line'}
        }
        expected = "A: ''  # one-line two-line.\n"
        self.assert_skeleton_equals(
            'yaml-input', expected, input_shape=input_shape
        )

    def test_generate_yaml_input_ignores_line_separator_char(self):
        input_shape = {
            'A': {'type': 'string', 'documentation': 'one-line\u2028two-line'}
        }
        expected = "A: ''  # one-line two-line.\n"
        self.assert_skeleton_equals(
            'yaml-input', expected, input_shape=input_shape
        )

    def test_generate_yaml_input_removes_xml_tags(self):
        input_shape = {'A': {'type': 'string', 'documentation': '<p>text</p>'}}
        expected = "A: ''  # text.\n"
        self.assert_skeleton_equals(
            'yaml-input', expected, input_shape=input_shape
        )

    def test_generate_yaml_input_can_handle_broken_xml(self):
        input_shape = {'A': {'type': 'string', 'documentation': '<p>text'}}
        expected = "A: ''  # <p>text.\n"
        self.assert_skeleton_equals(
            'yaml-input', expected, input_shape=input_shape
        )

    def test_generate_yaml_input_uses_only_content_before_first_period(self):
        input_shape = {
            'A': {
                'type': 'string',
                'documentation': 'First sentence. Second Sentence',
            }
        }
        expected = "A: ''  # First sentence.\n"
        self.assert_skeleton_equals(
            'yaml-input', expected, input_shape=input_shape
        )

    def test_generate_yaml_input_uses_only_content_before_first_colon(self):
        input_shape = {
            'A': {
                'type': 'string',
                'documentation': 'First sentence: * Option 1 * Option 2',
            }
        }
        expected = "A: ''  # First sentence.\n"
        self.assert_skeleton_equals(
            'yaml-input', expected, input_shape=input_shape
        )

    def test_generate_yaml_input_includes_enums_in_comments(self):
        input_shape = {
            'A': {
                'type': 'string',
                'documentation': 'First sentence.',
                'enum': ['ENUM1'],
            }
        }
        expected = "A: ENUM1  # First sentence. Valid values are: ENUM1.\n"
        self.assert_skeleton_equals(
            'yaml-input', expected, input_shape=input_shape
        )



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import io
import locale
import os
import signal

import pytest

from awscli.compat import (
    compat_open,
    compat_shell_quote,
    ensure_text_type,
    get_popen_kwargs_for_pager_cmd,
    getpreferredencoding,
    ignore_user_entered_signals,
    set_preferred_output_encoding,
    validate_preferred_output_encoding,
)
from awscli.testutils import FileCreator, mock, skip_if_windows, unittest


class TestEnsureText(unittest.TestCase):
    def test_string(self):
        value = 'foo'
        response = ensure_text_type(value)
        self.assertIsInstance(response, str)
        self.assertEqual(response, 'foo')

    def test_binary(self):
        value = b'bar'
        response = ensure_text_type(value)
        self.assertIsInstance(response, str)
        self.assertEqual(response, 'bar')

    def test_unicode(self):
        value = 'baz'
        response = ensure_text_type(value)
        self.assertIsInstance(response, str)
        self.assertEqual(response, 'baz')

    def test_non_ascii(self):
        value = b'\xe2\x9c\x93'
        response = ensure_text_type(value)
        self.assertIsInstance(response, str)
        self.assertEqual(response, '\u2713')

    def test_non_string_or_bytes_raises_error(self):
        value = 500
        with self.assertRaises(ValueError):
            ensure_text_type(value)


@pytest.mark.parametrize(
    "input_string, expected_output",
    (
        ('', '""'),
        ('"', '\\"'),
        ('\\', '\\'),
        ('\\a', '\\a'),
        ('\\\\', '\\\\'),
        ('\\"', '\\\\\\"'),
        ('\\\\"', '\\\\\\\\\\"'),
        ('foo bar', '"foo bar"'),
        ('foo\tbar', '"foo\tbar"'),
    ),
)
def test_compat_shell_quote_windows(input_string, expected_output):
    assert compat_shell_quote(input_string, "win32") == expected_output


@pytest.mark.parametrize(
    "input_string, expected_output",
    (
        ('', '""'),
        ('foo', 'foo'),
        ('foo bar', '"foo bar"'),
        ('foo\tbar', '"foo\tbar"'),
        ('"', '"\\""'),
        ('\\', '\\'),
        ('\\a', '\\a'),
        ('\\\\', '\\\\'),
        ('\\"', '"\\\\\\""'),
        ('foo&bar', '"foo&bar"'),
        ('foo|bar', '"foo|bar"'),
        ('foo>bar', '"foo>bar"'),
        ('foo<bar', '"foo<bar"'),
        ('foo^bar', '"foo^bar"'),
        ('foo(bar)', '"foo(bar)"'),
        ('foo,bar', '"foo,bar"'),
        ('foo;bar', '"foo;bar"'),
        ('foo=bar', '"foo=bar"'),
        ('foo!bar', '"foo!bar"'),
        ('foo%PATH%bar', 'foo%PATH%bar'),
        ('foo\\', 'foo\\'),
        ('foo bar\\', '"foo bar\\\\"'),
    ),
)
def test_compat_shell_quote_windows_for_cmd_exe(input_string, expected_output):
    assert (
        compat_shell_quote(input_string, "win32", shell=True)
        == expected_output
    )


@pytest.mark.parametrize(
    "input_string, expected_output",
    (
        ('', "''"),
        ('*', "'*'"),
        ('foo', 'foo'),
        ('foo bar', "'foo bar'"),
        ('foo\tbar', "'foo\tbar'"),
        ('foo\nbar', "'foo\nbar'"),
        ("foo'bar", '\'foo\'"\'"\'bar\''),
    ),
)
def test_comat_shell_quote_linux(input_string, expected_output):
    assert compat_shell_quote(input_string, "linux2") == expected_output


@pytest.mark.parametrize(
    "input_string, expected_output",
    (
        ('', "''"),
        ('*', "'*'"),
        ('foo', 'foo'),
        ('foo bar', "'foo bar'"),
        ('foo\tbar', "'foo\tbar'"),
        ('foo\nbar', "'foo\nbar'"),
        ("foo'bar", '\'foo\'"\'"\'bar\''),
    ),
)
def test_comat_shell_quote_darwin(input_string, expected_output):
    assert compat_shell_quote(input_string, "darwin") == expected_output


class TestGetPopenPagerCmd(unittest.TestCase):
    @mock.patch('awscli.compat.is_windows', True)
    @mock.patch('awscli.compat.default_pager', 'more')
    def test_windows(self):
        kwargs = get_popen_kwargs_for_pager_cmd()
        self.assertEqual({'args': 'more', 'shell': True}, kwargs)

    @mock.patch('awscli.compat.is_windows', True)
    @mock.patch('awscli.compat.default_pager', 'more')
    def test_windows_with_specific_pager(self):
        kwargs = get_popen_kwargs_for_pager_cmd('less -R')
        self.assertEqual({'args': 'less -R', 'shell': True}, kwargs)

    @mock.patch('awscli.compat.is_windows', False)
    @mock.patch('awscli.compat.default_pager', 'less -R')
    def test_non_windows(self):
        kwargs = get_popen_kwargs_for_pager_cmd()
        self.assertEqual({'args': ['less', '-R']}, kwargs)

    @mock.patch('awscli.compat.is_windows', False)
    @mock.patch('awscli.compat.default_pager', 'less -R')
    def test_non_windows_specific_pager(self):
        kwargs = get_popen_kwargs_for_pager_cmd('more')
        self.assertEqual({'args': ['more']}, kwargs)


class TestIgnoreUserSignals(unittest.TestCase):
    @skip_if_windows("These signals are not supported for windows")
    def test_ignore_signal_sigint(self):
        with ignore_user_entered_signals():
            try:
                os.kill(os.getpid(), signal.SIGINT)
            except KeyboardInterrupt:
                self.fail(
                    'The ignore_user_entered_signals context '
                    'manager should have ignored'
                )

    @skip_if_windows("These signals are not supported for windows")
    def test_ignore_signal_sigquit(self):
        with ignore_user_entered_signals():
            self.assertEqual(signal.getsignal(signal.SIGQUIT), signal.SIG_IGN)
            os.kill(os.getpid(), signal.SIGQUIT)

    @skip_if_windows("These signals are not supported for windows")
    def test_ignore_signal_sigtstp(self):
        with ignore_user_entered_signals():
            self.assertEqual(signal.getsignal(signal.SIGTSTP), signal.SIG_IGN)
            os.kill(os.getpid(), signal.SIGTSTP)


class TestGetPreferredEncoding(unittest.TestCase):
    @mock.patch.dict(os.environ, {'AWS_CLI_FILE_ENCODING': 'cp1252'})
    def test_getpreferredencoding_with_env_var(self):
        encoding = getpreferredencoding()
        self.assertEqual(encoding, 'cp1252')

    @mock.patch('locale.setlocale', side_effect=['POSIX', 'C'])
    def test_getpreferredencoding_wo_env_var_with_ctype_posix(self, *args):
        encoding = getpreferredencoding()
        self.assertEqual(encoding, 'UTF-8')
        encoding = getpreferredencoding()
        self.assertEqual(encoding, 'UTF-8')

    @mock.patch('locale.setlocale', return_value='English_United States.1252')
    @mock.patch('locale.getpreferredencoding')
    def test_runs_locale_getpreferredencoding_wo_env_var_and_posix(
        self, getprefedencoding, *args
    ):
        getpreferredencoding()
        getprefedencoding.assert_called_once_with()


class TestCompatOpenWithAccessPermissions(unittest.TestCase):
    def setUp(self):
        self.files = FileCreator()

    def tearDown(self):
        self.files.remove_all()

    @skip_if_windows('Permissions tests only supported on mac/linux')
    def test_can_create_file_with_acess_permissions(self):
        file_path = os.path.join(self.files.rootdir, "foo_600.txt")
        with compat_open(file_path, access_permissions=0o600, mode='w') as f:
            f.write('bar')
        self.assertEqual(os.stat(file_path).st_mode & 0o777, 0o600)

    def test_not_override_existing_file_access_permissions(self):
        file_path = os.path.join(self.files.rootdir, "foo.txt")
        with open(file_path, mode='w') as f:
            f.write('bar')
        expected_st_mode = os.stat(file_path).st_mode

        with compat_open(file_path, access_permissions=0o600, mode='w') as f:
            f.write('bar')
        self.assertEqual(os.stat(file_path).st_mode, expected_st_mode)


@pytest.mark.parametrize(
    'env_vars, expected_encoding',
    [
        ({}, 'cp1252'),
        ({'AWS_CLI_OUTPUT_ENCODING': 'UTF-8'}, 'UTF-8'),
        ({'PYTHONUTF8': '1'}, 'UTF-8'),
    ],
)
def test_set_preferred_output_encoding(env_vars, expected_encoding):
    stdout_b = io.BytesIO()
    stdout = io.TextIOWrapper(stdout_b, encoding="cp1252")

    with mock.patch.dict(os.environ, env_vars):
        set_preferred_output_encoding(stdout)

    assert stdout.encoding == expected_encoding


def test_validate_preferred_output_encoding():
    with mock.patch.dict(os.environ, {'AWS_CLI_OUTPUT_ENCODING': 'invalid'}):
        with pytest.raises(ValueError):
            validate_preferred_output_encoding()



#!/usr/bin/env python
# Copyright 2012-2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import base64
import contextlib
import io
import platform
import sys

from botocore.compat import json

from awscli.compat import StringIO, get_stdout_text_writer
from awscli.formatter import JSONFormatter
from awscli.testutils import (
    BaseAWSCommandParamsTest,
    mock,
    skip_if_windows,
    unittest,
)


class TestGetPasswordData(BaseAWSCommandParamsTest):
    COMMAND = 'iam add-user-to-group --group-name foo --user-name bar'

    def setUp(self):
        super(TestGetPasswordData, self).setUp()
        self.parsed_response = {}

    def test_empty_dict_response_prints_nothing(self):
        # This is the default response, but we want to be explicit
        # that we're returning an empty dict.
        self.parsed_response = {}
        stdout = self.run_cmd(self.COMMAND, expected_rc=0)[0]
        self.assertEqual(stdout, '')

    def test_empty_list_prints_list(self):
        self.parsed_response = []
        stdout = self.run_cmd(self.COMMAND, expected_rc=0)[0]
        self.assertEqual(stdout, '[]\n')

    def test_empty_string_prints_nothing(self):
        self.parsed_response = ''
        stdout = self.run_cmd(self.COMMAND, expected_rc=0)[0]
        self.assertEqual(stdout, '""\n')


class TestListUsers(BaseAWSCommandParamsTest):
    def setUp(self):
        super(TestListUsers, self).setUp()
        self.parsed_response = {
            'Users': [
                {
                    "UserName": "testuser-50",
                    "Path": "/",
                    "CreateDate": "2013-02-12T19:08:52Z",
                    "UserId": "EXAMPLEUSERID",
                    "Arn": "arn:aws:iam::12345:user/testuser1",
                },
                {
                    "UserName": "testuser-51",
                    "Path": "/",
                    "CreateDate": "2012-10-14T23:53:39Z",
                    "UserId": "EXAMPLEUSERID",
                    "Arn": "arn:aws:iam::123456:user/testuser2",
                },
            ]
        }

    def test_json_response(self):
        output = self.run_cmd('iam list-users', expected_rc=0)[0]
        parsed_output = json.loads(output)
        self.assertIn('Users', parsed_output)
        self.assertEqual(len(parsed_output['Users']), 2)
        self.assertEqual(
            sorted(parsed_output['Users'][0].keys()),
            ['Arn', 'CreateDate', 'Path', 'UserId', 'UserName'],
        )

    def test_jmespath_json_response(self):
        jmespath_query = 'Users[*].UserName'
        output = self.run_cmd(
            'iam list-users --query %s' % jmespath_query, expected_rc=0
        )[0]
        parsed_output = json.loads(output)
        self.assertEqual(parsed_output, ['testuser-50', 'testuser-51'])

    def test_zero_value_is_printed(self):
        # Even though the integer 0 is false-like, we still
        # should be printing it to stdout if a jmespath query
        # evalutes to 0.
        jmespath_query = '`0`'
        output = self.run_cmd(
            'iam list-users --query %s' % jmespath_query, expected_rc=0
        )[0]
        self.assertEqual(output, '0\n')

    def test_unknown_output_type_from_env_var(self):
        # argparse already handles the case with a bad --output
        # specified on the CLI, we need to verify that a bad
        # output format from the env var still gives an error.
        self.environ['AWS_DEFAULT_OUTPUT'] = 'bad-output-type'
        self.run_cmd('iam list-users', expected_rc=255)

    @skip_if_windows('Encoding tests only supported on mac/linux')
    def test_json_prints_unicode_chars(self):
        self.parsed_response['Users'][1]['UserId'] = '\u2713'
        output = self.run_cmd('iam list-users', expected_rc=0)[0]
        with mock.patch('sys.stdout', StringIO()) as f:
            out = get_stdout_text_writer()
            out.write('\u2713')
            expected = f.getvalue()
        # We should not see the '\u<hex>' for of the unicode character.
        # It should be encoded into the default encoding.
        self.assertNotIn('\\u2713', output)
        self.assertIn(expected, output)


class TestFormattersHandleClosedPipes(unittest.TestCase):
    def test_fully_buffered_handles_io_error(self):
        args = mock.Mock(query=None)
        operation = mock.Mock(can_paginate=False)
        response = '{"Foo": "Bar"}'
        fake_closed_stream = mock.Mock(spec=StringIO)
        fake_closed_stream.flush.side_effect = IOError
        formatter = JSONFormatter(args)
        formatter('command_name', response, stream=fake_closed_stream)
        # We should not have let the IOError propogate, but
        # we still should have called the flush() on the
        # stream.
        fake_closed_stream.flush.assert_called_with()


class TestBinaryData(unittest.TestCase):
    def test_binary_data_gets_base64_encoded(self):
        args = mock.Mock(query=None)
        raw_bytes = b'foo'
        response = {'BinaryValue': raw_bytes}
        stdout_b = io.BytesIO()
        stdout = io.TextIOWrapper(stdout_b, newline='\n')
        formatter = JSONFormatter(args)

        with contextlib.redirect_stdout(stdout):
            formatter('command-name', response, sys.stdout)
            stdout.flush()

        assert (
            stdout_b.getvalue()
            == (
                '{\n'
                f'    "BinaryValue": "{base64.b64encode(raw_bytes).decode("utf-8")}"\n'
                '}\n'
            ).encode()
        )



# Copyright 2025 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import pytest
from botocore.args import PRIORITY_ORDERED_SUPPORTED_PROTOCOLS
from botocore.session import get_session


def _multi_protocol_test_cases():
    session = get_session()
    loader = session.get_component('data_loader')
    services = loader.list_available_services('service-2')
    multi_protocol_services = []
    supported_protocols = []
    for service in services:
        service_model = session.get_service_model(service)
        if 'protocols' in service_model.metadata:
            multi_protocol_services.append(service)
            supported_protocols.append(
                service_model.metadata.get('protocols', [])
            )
    return list(zip(multi_protocol_services, supported_protocols))


def _single_protocol_test_cases():
    session = get_session()
    loader = session.get_component('data_loader')
    services = loader.list_available_services('service-2')
    single_protocol_services = []
    supported_protocol = []
    for service in services:
        service_model = session.get_service_model(service)
        if 'protocols' not in service_model.metadata:
            single_protocol_services.append(service)
            supported_protocol.append(service_model.metadata.get('protocol'))
    return list(zip(single_protocol_services, supported_protocol))


@pytest.mark.validates_models
@pytest.mark.parametrize(
    "service_name, supported_protocols",
    _multi_protocol_test_cases(),
)
def test_services_with_protocols_trait_have_supported_protocol(
    service_name, supported_protocols, record_property
):
    message = f"No protocols supported for service {service_name}"
    # Store the service name in PyTest custom properties
    record_property('aws_service', service_name)
    assert any(
        protocol in PRIORITY_ORDERED_SUPPORTED_PROTOCOLS
        for protocol in supported_protocols
    ), message


@pytest.mark.validates_models
@pytest.mark.parametrize(
    "service_name, supported_protocol",
    _single_protocol_test_cases(),
)
def test_services_without_protocols_trait_have_supported_protocol(
    service_name, supported_protocol, record_property
):
    message = f"Service protocol not supported for {service_name}"
    # Store the service name in PyTest custom properties
    record_property('aws_service', service_name)
    assert supported_protocol in PRIORITY_ORDERED_SUPPORTED_PROTOCOLS, message
